import importlib
import os
import re
import secrets
import string
import sys
import time
import types
from collections import deque
from io import StringIO
from itertools import tee, filterfalse, count
from pathlib import Path
from typing import Union

import blosc


class LazyLoader(types.ModuleType):
    """Lazily import a module, mainly to avoid pulling in large dependencies.
    `tensorflow`, and `pytorch` are examples of modules that are large and not always
    needed, and this allows them to only be loaded when they are used.
    """

    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        super(LazyLoader, self).__init__(name)

    def _load(self):
        """Load the module and insert it into the parent's globals.
        """
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


def is_64bits():
    """bool indicating if running on atleast a 64 bit machine
    """
    return sys.maxsize > 2 ** 32


def set_blosc_nthreads() -> int:  # pragma: no cover
    """set the blosc library to two less than the core count on the system.

    If less than 2 cores are ncores-2, we set the value to two.

    Returns
    -------
    int
        ncores blosc will use on the system
    """
    nCores = blosc.detect_number_of_cores()
    if nCores == 1:
        nUsed = 1
    elif nCores == 2:
        nUsed = 2
    elif nCores <= 4:
        nUsed = nCores - 1
    else:
        nUsed = nCores - 2
    blosc.set_nthreads(nUsed)
    return nUsed


def random_string(
    n: int = 8,
    *, _ALPHABET=''.join([string.ascii_lowercase, string.digits])
) -> str:
    """Generate a random string of lowercase ascii letters and digits.

    Parameters
    ----------
    n: int, optional
        The number of characters which the output string will have. Default=8
    """
    token = [secrets.choice(_ALPHABET) for i in range(n)]
    return ''.join(token)


_SuitableCharRE = re.compile(r'[\w\.\-\_]+\Z', flags=re.ASCII)


def is_suitable_user_key(key: Union[str, int]) -> bool:
    """Checks if only alpha-numeric ascii chars or ['.', '-' '_'] (no whitespace)

    Necessary because python 3.6 does not have a str.isascii() method. In
    addition, checks that all input keys are less than 64 characters long.

    Parameters
    ----------
    key : Union[str, int]
        string to check if it contains only ascii characters

    Returns
    -------
    bool
        True if only ascii characters in the string, else False.
    """
    try:
        if isinstance(key, int) and (key >= 0):
            str_data = str(key)
        elif isinstance(key, str):
            str_data = str(key)
        else:
            raise TypeError
        if len(str_data) > 64:
            return False
        return bool(_SuitableCharRE.match(str_data))
    except TypeError:
        return False


def is_ascii(str_data: str) -> bool:
    """Checks if string contains only ascii chars.

    Necessary because python 3.6 does not have a str.isascii() method.

    Parameters
    ----------
    str_data : str
        string to check if it contains only ascii characters

    Returns
    -------
    bool
        True if only ascii characters in the string, else False.
    """
    try:
        str_data.encode('ascii')
    except (UnicodeEncodeError, AttributeError):
        return False
    return True


def valfilter(predicate, d, factory=dict):
    """ Filter items in dictionary by values that are true.

    >>> iseven = lambda x: x % 2 == 0
    >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
    >>> valfilter(iseven, d)
    {1: 2, 3: 4}

    See Also:
        valfilterfalse
    """
    rv = factory()
    for k, v in d.items():
        if predicate(v):
            rv[k] = v
    return rv


def valfilterfalse(predicate, d, factory=dict):
    """ Filter items in dictionary by values which are false.

    >>> iseven = lambda x: x % 2 == 0
    >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
    >>> valfilterfalse(iseven, d)
    {2: 3, 4: 5}

    See Also:
        valfilter
    """
    rv = factory()
    for k, v in d.items():
        if not predicate(v):
            rv[k] = v
    return rv


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def unique_everseen(iterable, key=None):
    """List unique elements, preserving order. Remember all elements ever seen.

    >>> list(unique_everseen('AAAABBBCCDAABBB'))
    ['A', 'B', 'C', 'D']
    >>> list(unique_everseen('ABBCcAD', str.lower))
    ['A', 'B', 'C', 'D']
    """
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def ilen(iterable):
    """Return the number of items in *iterable*.

        >>> ilen(x for x in range(1000000) if x % 3 == 0)
        333334
        >>> it = iter([0, 1, 2, False])
        >>> ilen(it)
        4

    This consumes the iterable, so handle with care.
    """
    counter = count()
    deque(zip(iterable, counter), maxlen=0)
    return next(counter)


def find_next_prime(N: int) -> int:
    """Find next prime >= N

    Parameters
    ----------
    N : int
        Starting point to find the next prime >= N.

    Returns
    -------
    int
        the next prime found after the number N
    """
    def is_prime(n):
        if n % 2 == 0:
            return False
        i = 3
        while i * i <= n:
            if n % i != 0:
                i += 2
            else:
                return False
        return True

    if N < 3:
        return 2
    if N % 2 == 0:
        N += 1
    for n in range(N, 2 * N, 2):
        if is_prime(n):
            return n


def file_size(p: Path) -> int:  # pragma: no cover
    """Query the file size of a specific file

    Parameters
    ----------
    p : Path
        path to a file that exists on disk.

    Raises
    ------
    FileNotFoundError
        if the file does not exist

    Returns
    -------
    int
        nbytes the file consumes on disk.
    """
    if not p.is_file():
        err = f'Cannot query size of: {str(p)}. File does not exist'
        raise FileNotFoundError(err)
    return p.stat().st_size


def folder_size(p: Path, *, recurse: bool = False) -> int:
    """size of all files in a folder.

    Default is to not include subdirectories. Set "recurse=True"
    to enable recursive calculation.

    Parameters
    ----------
    p : Path
        path to the repository on disk.
    recurse : bool, kwarg-only
        to calculate the full size of the repo (Default value = False)

    Returns
    -------
    int
        number of bytes used up in the repo_path
    """
    total = 0
    for entry in p.iterdir():
        if entry.is_file() and not entry.is_symlink():
            total += entry.stat().st_size
        elif recurse and entry.is_dir() and not entry.is_symlink():
            total += folder_size(entry.resolve(), recurse=True)
    return total


def is_valid_directory_path(p: Path) -> Path:
    """Check if path is directory which user has write permission to.

    Parameters
    ----------
    p : Path
        path to some location on disk

    Returns
    -------
    Path
        If successful, the path with any user constructions expanded
        (ie. `~/somedir` -> `/home/foo/somedir`)

    Raises
    ------
    TypeError
        If the provided path argument is not a pathlike object
    NotADirectoryError
        If the path does not exist, or is not a directory on disk
    PermissionError
        If the user does not have write access to the specified path
    """
    if not isinstance(p, Path):
        msg = f'Path arg `p`: {p} of type: {type(p)} is not valid path specifier'
        raise TypeError(msg)

    usr_path = p.expanduser().resolve(strict=True)

    if not usr_path.is_dir():
        msg = f'Path arg `p`: {p} is not a directory.'
        raise NotADirectoryError(msg)
    if not os.access(str(usr_path), os.W_OK):  # pragma: no cover
        msg = f'User does not have permission to write to directory path: {p}'
        raise PermissionError(msg)

    return usr_path


# ----------------- human & machine nbytes ------------------------------------


def format_bytes(n: int) -> str:
    """ Format bytes as text
    >>> format_bytes(1)
    '1.00 B'
    >>> format_bytes(1234)
    '1.23 kB'
    >>> format_bytes(12345678)
    '12.35 MB'
    >>> format_bytes(1234567890)
    '1.23 GB'
    >>> format_bytes(1234567890000)
    '1.23 TB'
    >>> format_bytes(1234567890000000)
    '1.23 PB'
    """
    for x in ['B', 'kB', 'MB', 'GB', 'TB', 'PB']:
        if n < 1000.0:
            return "%3.2f %s" % (n, x)
        n /= 1000.0


_byte_sizes = {
    'kb': 1000,
    'mb': 1000000,
    'gb': 1000000000,
    'tb': 1000000000000,
    'pb': 1000000000000000,
    'kib': 1024,
    'mib': 1048576,
    'gib': 1073741824,
    'tib': 1099511627776,
    'pib': 1125899906842624,
    'b': 1,
    '': 1,
    'k': 1000,
    'm': 1000000,
    'g': 1000000000,
    't': 1000000000000,
    'p': 1000000000000000,
    'ki': 1024,
    'mi': 1048576,
    'gi': 1073741824,
    'ti': 1099511627776,
    'pi': 1125899906842624
}


def parse_bytes(s: str) -> int:
    """ Parse byte string to numbers
    >>> parse_bytes('100')
    100
    >>> parse_bytes('100 MB')
    100000000
    >>> parse_bytes('100M')
    100000000
    >>> parse_bytes('5kB')
    5000
    >>> parse_bytes('5.4 kB')
    5400
    >>> parse_bytes('1kiB')
    1024
    >>> parse_bytes('1e6')
    1000000
    >>> parse_bytes('1e6 kB')
    1000000000
    >>> parse_bytes('MB')
    1000000
    """
    s = s.replace(' ', '').lower()
    s = f'1{s}' if not s[0].isdigit() else s
    for i in range(len(s) - 1, -1, -1):
        if not s[i].isalpha():
            break

    n = float(s[:i + 1])
    mult = _byte_sizes[s[i + 1:]]
    return int(n * mult)


def readme_contents(user_name: str, user_email: str) -> StringIO:
    """Create the Hangar README.txt contents used to fill out file on repo initialization

    Parameters
    ----------
    user_name : str
        name of the user initializing the repository on the machine.
    user_email : str
        email of the user initializing the repository on the machine.

    Returns
    -------
    StringIO
        Buffered string text ready to be sent to a file writer.
    """
    from . import __version__
    from .constants import DIR_HANGAR

    buf = StringIO()
    buf.write(f'This directory has been used to initialize a Hangar Repository\n')
    buf.write(f'\n')
    buf.write(f'This repository was initialized by:\n')
    buf.write(f'    User Name:        {user_name}\n')
    buf.write(f'    User Email:       {user_email}\n')
    buf.write(f'    Creation Time:    {time.asctime(time.gmtime())} UTC\n')
    buf.write(f'    Software Version: {__version__}\n')
    buf.write(f'\n')
    buf.write(f'NOTE: The repository may have been updated to work with newer Hangar versions\n')
    buf.write(f'since initialization.\n')
    buf.write(f'\n')
    buf.write(f'Do not modify the contents of this `{DIR_HANGAR}` folder under any circumstances.\n')
    buf.write(f'The contents are not meant to be understood by humans. Doing so will result\n')
    buf.write(f'in data loss / corruption.\n')
    buf.write(f'\n')
    buf.write(f'The project homepage can be found at: https://github.com/tensorwerk/hangar-py/ \n')
    buf.write(f'Documention is available at: https://hangar-py.readthedocs.io/en/latest/ \n')
    buf.write(f'\n')
    buf.write(f'NOTE: If this Repository has been initialized in a directory under traditional\n')
    buf.write(f'version control systems, please add `{DIR_HANGAR}` as an ignored directory path.\n')
    buf.write(f'Failure to do so will result in undesireable performance of version control\n')
    buf.write(f'systems meant for text/code such as Git, Mercurial, Subversion, etc.\n')

    return buf



import json
from hashlib import blake2b
from itertools import cycle
from random import randint
from time import perf_counter, sleep
from typing import Union, NamedTuple, Tuple, Iterable

import blosc

from ..constants import (
    CMT_DIGEST_JOIN_KEY,
    CMT_KV_JOIN_KEY,
    CMT_REC_JOIN_KEY,
    K_BRANCH,
    K_HEAD,
    K_REMOTES,
    K_VERSION,
    K_WLOCK,
    SEP_CMT,
    SEP_KEY,
    WLOCK_SENTINAL,
)
from .._version import parse as version_parse
from .._version import Version


cycle_list = (str(c).rjust(4, '0') for c in range(9_999))
NAME_CYCLER = cycle(cycle_list)
RANDOM_NAME_SEED = str(randint(0, 999_999_999)).rjust(0, '0')
perf_counter()  # call to init monotonic start point


def generate_sample_name() -> str:
    ncycle = next(NAME_CYCLER)
    if ncycle == '0000':
        sleep(0.001)

    sec, subsec = str(perf_counter()).split('.')
    name = f'{RANDOM_NAME_SEED}{sec.rjust(6, "0")}{subsec.ljust(9, "0")}{ncycle}'
    return name


"""
Parsing functions used to deal with repository state The parsers defined in this
section handle repo/branch records.

Methods working with repository version specifiers
--------------------------------------------------
"""


def repo_version_raw_spec_from_raw_string(v_str: str) -> Version:
    """Convert from user facing string representation to Version object
    """
    return version_parse(v_str)


# ------------------------- db version key is fixed -----------------


def repo_version_db_key() -> bytes:
    """The db formated key which version information can be accessed at

    Returns
    -------
    bytes
        db formatted key to use to get/set the repository software version.
    """
    return K_VERSION.encode()


# ------------------------ raw -> db --------------------------------


def repo_version_db_val_from_raw_val(v_spec: Version) -> bytes:
    """determine repository version db specifier from version spec.

    Parameters
    ----------
    v_spec : Version
        This class abstracts handling of a project’s versions. A Version
        instance is comparison aware and can be compared and sorted using the
        standard Python interfaces.

    Returns
    -------
    bytes
        db formatted specification of version
    """
    return str(v_spec).encode()


# ---------------------------- db -> raw ----------------------------


def repo_version_raw_val_from_db_val(db_val: bytes) -> Version:
    """determine software version of hangar repository is written for.

    Parameters
    ----------
    db_val : bytes
        db formatted specification of version string

    Returns
    -------
    Version
        This class abstracts handling of a project’s versions. A Version
        instance is comparison aware and can be compared and sorted using the
        standard Python interfaces.
    """
    db_str = db_val.decode()
    return version_parse(db_str)


"""
Methods working with writer HEAD branch name
--------------------------------------------
"""

# --------------------- db HEAD key is fixed ------------------------


def repo_head_db_key() -> bytes:
    """db_key of the head staging branch name.

    Returns
    -------
    bytestring
        lmdb key to query while looking up the head staging branch name
    """
    return K_HEAD.encode()


# --------------------- raw -> db -----------------------------------


def repo_head_db_val_from_raw_val(branch_name: str) -> bytes:
    return f'{K_BRANCH}{branch_name}'.encode()


# --------------------- db -> raw -----------------------------------

def repo_head_raw_val_from_db_val(db_val: bytes) -> str:
    return db_val.decode()[len(K_BRANCH):]


"""
Methods working with branch names / head commit values
------------------------------------------------------
"""

# ---------------------- raw -> db --------------------------------


def repo_branch_head_db_key_from_raw_key(branch_name: str) -> bytes:
    return f'{K_BRANCH}{branch_name}'.encode()


def repo_branch_head_db_val_from_raw_val(commit_hash: str) -> bytes:
    return f'{commit_hash}'.encode()


# ---------------- db -> raw -----------------------------------


def repo_branch_head_raw_key_from_db_key(db_key: bytes) -> str:
    return db_key.decode()[len(K_BRANCH):]


def repo_branch_head_raw_val_from_db_val(db_val: bytes) -> str:
    try:
        commit_hash = db_val.decode()
    except AttributeError:
        commit_hash = ''
    return commit_hash


"""
Methods working with writer lock key/values
-------------------------------------------
"""

# ------------------- db key for lock is fixed -------------------


def repo_writer_lock_db_key() -> bytes:
    return K_WLOCK.encode()


def repo_writer_lock_sentinal_db_val() -> bytes:
    return WLOCK_SENTINAL.encode()


def repo_writer_lock_force_release_sentinal() -> str:
    return 'FORCE_RELEASE'


# ------------------------- raw -> db ------------------------------


def repo_writer_lock_db_val_from_raw_val(lock_uuid: str) -> bytes:
    return f'{lock_uuid}'.encode()


# -------------------------- db -> raw ------------------------------


def repo_writer_lock_raw_val_from_db_val(db_val: bytes) -> str:
    return db_val.decode()


# -------------------- Remote Work --------------------------------------------


def remote_db_key_from_raw_key(remote_name: str) -> bytes:
    """Get the remote db key val for a remote name

    Parameters
    ----------
    remote_name : str
        name of the remote location

    Returns
    -------
    bytes
        db key allowing access to address value at the name of the remote
    """
    return f'{K_REMOTES}{remote_name}'.encode()


def remote_raw_key_from_db_key(db_key: bytes, *, _SPLT=len(K_REMOTES)) -> str:
    """Get the remote name from a remote db key

    Parameters
    ----------
    db_key : bytes
        db key of the remote

    Returns
    -------
    str
        name of the remote
    """
    return db_key.decode()[_SPLT:]


def remote_db_val_from_raw_val(grpc_address: str) -> bytes:
    """Format a remote db value from it's grpc address string

    Parameters
    ----------
    grpc_address : str
        IP:PORT where the grpc server can be accessed

    Returns
    -------
    bytes
        formated representation of the grpc address suitable for storage in lmdb.
    """
    return grpc_address.encode()


def remote_raw_val_from_db_val(db_val: bytes) -> str:
    """Retrieve the address where a grpc server is running from a remote db value


    Parameters
    ----------
    db_val : bytes
        db value assigned to the desired remote name

    Returns
    -------
    str
        IP:PORT where the grpc server can be accessed.
    """
    return db_val.decode()


"""
Commit Parsing Methods
-----------------------

The parsers defined in this section handle commit (ref) records
"""


class CommitAncestorSpec(NamedTuple):
    is_merge_commit: bool
    master_ancestor: str
    dev_ancestor: str


class CommitUserSpec(NamedTuple):
    commit_time: float
    commit_message: str
    commit_user: str
    commit_email: str


class DigestAndUserSpec(NamedTuple):
    digest: str
    user_spec: CommitUserSpec


class DigestAndAncestorSpec(NamedTuple):
    digest: str
    ancestor_spec: CommitAncestorSpec


class DigestAndBytes(NamedTuple):
    digest: str
    raw: bytes


class DigestAndDbRefs(NamedTuple):
    digest: str
    db_kvs: Union[Tuple, Tuple[Tuple[bytes, bytes]]]


def _hash_func(recs: bytes) -> str:
    """hash a tuple of db formatted k, v pairs.

    Parameters
    ----------
    recs : bytes
        tuple to calculate the joined digest of

    Returns
    -------
    str
        hexdigest of the joined tuple data
    """
    return blake2b(recs, digest_size=20).hexdigest()


def cmt_final_digest(parent_digest: str, spec_digest: str, refs_digest: str,
                     *, tcode: str = 'a') -> str:
    """Determine digest of commit based on digests of its parent, specs, and refs.

    Parameters
    ----------
    parent_digest : str
        digest of the parent value
    spec_digest : str
        digest of the user spec value
    refs_digest : str
        digest of the data record values
    tcode : str, optional, kwarg-only
        hash calculation type code. Included to allow future updates to change
        hashing algorithm, kwarg-only, by default '0'

    Returns
    -------
    str
        digest of the commit with typecode prepended by '{tcode}='.
    """
    if tcode == 'a':
        sorted_digests = sorted([parent_digest, spec_digest, refs_digest])
        joined_bytes = CMT_DIGEST_JOIN_KEY.join(sorted_digests).encode()
        rawDigest = _hash_func(joined_bytes)
        digest = f'a={rawDigest}'
    else:
        raise ValueError(
            f'Invalid commit reference type code {tcode}. If encountered during '
            f'normal operation, please report to hangar development team.')
    return digest


"""
Commit Parent (ancestor) Lookup methods
---------------------------------------
"""

# ------------------------- raw -> db -----------------------------------------


def commit_parent_db_key_from_raw_key(commit_hash: str) -> bytes:
    return f'{commit_hash}'.encode()


def commit_parent_db_val_from_raw_val(master_ancestor: str,
                                      dev_ancestor: str = '',
                                      is_merge_commit: bool = False) -> DigestAndBytes:
    if is_merge_commit:
        str_val = f'{master_ancestor}{SEP_CMT}{dev_ancestor}'
    else:
        str_val = f'{master_ancestor}'
    db_val = str_val.encode()
    digest = _hash_func(db_val)
    return DigestAndBytes(digest=digest, raw=db_val)


# ------------------------------- db -> raw -----------------------------------


def commit_parent_raw_key_from_db_key(db_key: bytes) -> str:
    return db_key.decode()


def commit_parent_raw_val_from_db_val(db_val: bytes) -> DigestAndAncestorSpec:
    """Parse the value of a commit's parent value to find it's ancestors

    Parameters
    ----------
    db_val : bytes
        Lmdb value of the commit parent field.

    Returns
    -------
    DigestAndAncestorSpec
        `digest` of data writen to disk and `ancestor_spec`, Namedtuple
        containing fields for `is_merge_commit`, `master_ancestor`, and
        `dev_ancestor`
    """
    parentValDigest = _hash_func(db_val)

    commit_str = db_val.decode()
    commit_ancestors = commit_str.split(SEP_CMT)
    if len(commit_ancestors) == 1:
        is_merge_commit = False
        master_ancestor = commit_ancestors[0]
        dev_ancestor = ''
    else:
        is_merge_commit = True
        master_ancestor = commit_ancestors[0]
        dev_ancestor = commit_ancestors[1]

    ancestorSpec = CommitAncestorSpec(is_merge_commit, master_ancestor, dev_ancestor)
    return DigestAndAncestorSpec(digest=parentValDigest, ancestor_spec=ancestorSpec)


"""
Commit reference key and values.
--------------------------------
"""


def commit_ref_db_key_from_raw_key(commit_hash: str) -> bytes:
    return f'{commit_hash}{SEP_KEY}ref'.encode()


def _commit_ref_joined_kv_digest(joined_db_kvs: Iterable[bytes]) -> str:
    """reproducibly calculate digest from iterable of joined record k/v pairs.

    First calculate the digest of each element in the input iterable. As these
    elements contain the record type (meta key, column name, sample key) as
    well as the data hash digest, any modification of any reference record will
    result in a different digest for that element. Then join all elements into
    single serialized bytestring.

    The output of this method is the hash digest of the serialized bytestring.

    Parameters
    ----------
    joined_db_kvs : Iterable[bytes]
        list or tuple of bytes where each element is the joining of kv pairs
        from the full commit references

    Returns
    -------
    str
        calculated digest of the commit ref record component
    """
    kv_digests = map(_hash_func, joined_db_kvs)
    joined_digests = CMT_DIGEST_JOIN_KEY.join(kv_digests).encode()
    res = _hash_func(joined_digests)
    return res


def commit_ref_db_val_from_raw_val(db_kvs: Iterable[Tuple[bytes, bytes]]) -> DigestAndBytes:
    """serialize and compress a list of db_key/db_value pairs for commit storage

    Parameters
    ----------
    db_kvs : Iterable[Tuple[bytes, bytes]]
        Iterable collection binary encoded db_key/db_val pairs.

    Returns
    -------
    DigestAndBytes
        `raw` serialized and compressed representation of the object. `digest`
        digest of the joined db kvs.
    """
    joined = tuple(map(CMT_KV_JOIN_KEY.join, db_kvs))
    refDigest = _commit_ref_joined_kv_digest(joined)
    pck = CMT_REC_JOIN_KEY.join(joined)
    raw = blosc.compress(pck, typesize=1, clevel=8, shuffle=blosc.NOSHUFFLE, cname='zstd')
    return DigestAndBytes(digest=refDigest, raw=raw)


def commit_ref_raw_val_from_db_val(commit_db_val: bytes) -> DigestAndDbRefs:
    """Load and decompress a commit ref db_val into python object memory.

    Parameters
    ----------
    commit_db_val : bytes
        Serialized and compressed representation of commit refs.

    Returns
    -------
    DigestAndDbRefs
        `digest` of the unpacked commit refs if desired for verification. `db_kvs`
        Iterable of binary encoded key/value pairs making up the repo state at the
        time of that commit. key/value pairs are already in sorted order.
    """
    uncomp_db_raw = blosc.decompress(commit_db_val)
    # if a commit has nothing in it (completely empty), the return from query == ()
    # the stored data is b'' from which the hash is calculated. We manually set these
    # values as the expected unpacking routine will not work correctly.
    if uncomp_db_raw == b'':
        refsDigest = _hash_func(b'')
        raw_db_kv_list = ()
    else:
        raw_joined_kvs_list = uncomp_db_raw.split(CMT_REC_JOIN_KEY)
        refsDigest = _commit_ref_joined_kv_digest(raw_joined_kvs_list)
        raw_db_kv_list = tuple(map(tuple, map(bytes.split, raw_joined_kvs_list)))

    return DigestAndDbRefs(digest=refsDigest, db_kvs=raw_db_kv_list)


"""
Commit spec reference keys and values
-------------------------------------
"""


def commit_spec_db_key_from_raw_key(commit_hash: str) -> bytes:
    return f'{commit_hash}{SEP_KEY}spec'.encode()


def commit_spec_db_val_from_raw_val(commit_time: float, commit_message: str,
                                    commit_user: str,
                                    commit_email: str) -> DigestAndBytes:
    """Serialize a commit specification from user values to a db store value

    Parameters
    ----------
    commit_time : float
        time since unix epoch that the commit was made
    commit_message : str
        user specified commit message to attach to the record
    commit_user : str
        globally configured user name of the repository committer
    commit_email : str
        globally configured user email of the repository committer

    Returns
    -------
    DigestAndBytes
        Two tuple containing ``digest`` and ``raw`` compressed binary encoded
        serialization of commit spec
    """
    spec_dict = {
        'commit_time': commit_time,
        'commit_message': commit_message,
        'commit_user': commit_user,
        'commit_email': commit_email,
    }

    db_spec_val = json.dumps(spec_dict, separators=(',', ':')).encode()
    digest = _hash_func(db_spec_val)
    comp_raw = blosc.compress(
        db_spec_val, typesize=8, clevel=9, shuffle=blosc.SHUFFLE, cname='zlib')
    return DigestAndBytes(digest=digest, raw=comp_raw)


def commit_spec_raw_val_from_db_val(db_val: bytes) -> DigestAndUserSpec:
    uncompressed_db_val = blosc.decompress(db_val)
    digest = _hash_func(uncompressed_db_val)
    commit_spec = json.loads(uncompressed_db_val)
    user_spec = CommitUserSpec(**commit_spec)
    return DigestAndUserSpec(digest=digest, user_spec=user_spec)



from .column import Columns, ModifierTypes
from .common import ColumnTxn
from .constructors import (
    generate_flat_column,
    generate_nested_column,
    column_type_object_from_schema
)

__all__ = (
    'Columns',
    'ModifierTypes',
    'generate_flat_column',
    'generate_nested_column',
    'column_type_object_from_schema',
    'ColumnTxn',
)



from typing import Dict, Iterable, Iterator, List, Set, Tuple, Union

import lmdb

from .column_parsers import (
    data_record_digest_val_from_db_val,
    dynamic_layout_data_record_db_start_range_key,
    dynamic_layout_data_record_from_db_key,
    schema_column_record_from_db_key,
    schema_db_range_key_from_column_unknown_layout,
    schema_record_count_start_range_key,
)
from .recordstructs import (
    FlatColumnDataKey,
    NestedColumnDataKey,
    DataRecordVal,
)
from ..txnctx import TxnRegister
from ..utils import ilen
from ..mixins import CursorRangeIterator

RawDataTuple = Tuple[Union[FlatColumnDataKey, NestedColumnDataKey], DataRecordVal]


class RecordQuery(CursorRangeIterator):

    def __init__(self, dataenv: lmdb.Environment):
        self._dataenv = dataenv

# ------------------ traversing the unpacked records ------------------------------------

    def _traverse_all_records(self) -> Iterator[Tuple[bytes, bytes]]:
        """Pull out all records in the database as a tuple of binary encoded

        Returns
        -------
        list of tuples of bytes
            list type stack of tuples with each db_key, db_val pair
        """
        try:
            datatxn = TxnRegister().begin_reader_txn(self._dataenv)
            with datatxn.cursor() as cursor:
                cursor.first()
                for db_kv in cursor.iternext(keys=True, values=True):
                    yield db_kv
        finally:
            TxnRegister().abort_reader_txn(self._dataenv)

    def _traverse_column_schema_records(self, keys: bool = True, values: bool = True
                                        ) -> Iterable[Union[Tuple[bytes], Tuple[bytes, bytes]]]:
        """Internal method to traverse all schema records and pull out k/v db pairs.

        Parameters
        ----------
        keys : bool, optional
            If True, yield metadata keys encountered, if False only values are returned.
            By default, True.
        values : bool, optional
            If True, yield metadata hash values encountered, if False only keys are returned.
            By default, True.

        Yields
        ------
        Iterable[Union[Tuple[bytes], Tuple[bytes, bytes]]]:
            db schema keys and db_values
        """
        startSchemaRangeKey = schema_record_count_start_range_key()
        try:
            datatxn = TxnRegister().begin_reader_txn(self._dataenv)
            yield from self.cursor_range_iterator(datatxn, startSchemaRangeKey, keys, values)
        finally:
            TxnRegister().abort_reader_txn(self._dataenv)

    def _traverse_column_data_records(self,
                                      column_name: str,
                                      *,
                                      keys: bool = True,
                                      values: bool = True) -> Iterable[Union[bytes, Tuple[bytes, bytes]]]:
        """Internal method to traverse column data records and get keys/db_values

        The column name is required because this method controls the cursor
        movement by first setting it's position on the column record count
        key, reading it's value "N" and then sequentially pulling records out of
        the db for N loops.

        Parameters
        ----------
        column_name : str
            name of the column to traverse records for.
        keys : bool, optional
            If True, yield metadata keys encountered, if False only values are returned.
            By default, True.
        values : bool, optional
            If True, yield metadata hash values encountered, if False only keys are returned.
            By default, True.

        Yields
        ------
        Iterable[Union[bytes, Tuple[bytes, bytes]]]:
            dict of db_key/db_values for each record traversed

        Raises
        ------
        KeyError
            if no column exists with the requested name.
        """
        try:
            datatxn = TxnRegister().begin_reader_txn(self._dataenv)
            schemaColumnRangeKey = schema_db_range_key_from_column_unknown_layout(column_name)
            with datatxn.cursor() as cur:
                if not cur.set_range(schemaColumnRangeKey):
                    raise KeyError(f'Traversal of commit references failed. '
                                   f'No column named `{column_name}` exists.')
                schemaColumnKey = cur.key()
            column_record = schema_column_record_from_db_key(schemaColumnKey)
            startRangeKey = dynamic_layout_data_record_db_start_range_key(column_record)
            yield from self.cursor_range_iterator(datatxn, startRangeKey, keys, values)
        finally:
            TxnRegister().abort_reader_txn(self._dataenv)

# ------------------------- process columns --------------------------------------------

    def column_names(self) -> List[str]:
        """Find all named columns in the checkout

        Returns
        -------
        List[str]
            list of all column names
        """
        recs = self._traverse_column_schema_records(keys=True, values=False)
        column_recs = map(schema_column_record_from_db_key, recs)
        return [x.column for x in column_recs]

    def column_count(self) -> int:
        """Return number of columns/schemas in the commit

        Returns
        -------
        int
            len of columns
        """
        return ilen(self._traverse_column_schema_records(keys=True, values=False))

    def data_hashes(self) -> List[str]:
        """Find all data hashes contained within all columns

        Note: this method does not deduplicate values

        Returns
        -------
        List[str]
            all hash values for all data pieces in the commit
        """
        all_hashes = []
        columns = self.column_names()
        for column in columns:
            recs = self._traverse_column_data_records(column, keys=False, values=True)
            data_rec = map(data_record_digest_val_from_db_val, recs)
            data_val_rec = [x.digest for x in data_rec]
            all_hashes.extend(data_val_rec)
        return all_hashes

# ------------------------ process column data records ----------------------

    def column_data_records(self, column_name: str) -> Iterable[RawDataTuple]:
        """Returns the raw data record key and record values for a specific column.

        Parameters
        ----------
        column_name : str
            name of the column to pull records for

        Yields
        ------
        Iterable[RawDataTuple]
            generator of key and value data record specs
        """
        for data_key, data_val in self._traverse_column_data_records(column_name):
            data_rec_key = dynamic_layout_data_record_from_db_key(data_key)
            data_rec_val = data_record_digest_val_from_db_val(data_val)
            yield (data_rec_key, data_rec_val)

    def column_data_hashes(self, column_name: str) -> Set[DataRecordVal]:
        """Find all data hashes contained within a particular column

        Note: this method does not remove any duplicates which may be present,
        if dedup is required, process it downstream

        Parameters
        ----------
        column_name : str
            name of the column to find the hashes contained in

        Returns
        -------
        Set[DataRecordVal]
            all hash values for all data pieces in the column
        """
        recs = self._traverse_column_data_records(column_name, keys=False, values=True)
        return set(map(data_record_digest_val_from_db_val, recs))

    def column_data_count(self, column_name: str) -> int:
        """Return the number of samples in an column with the provided name

        Parameters
        ----------
        column_name : str
            name of the column to query

        Returns
        -------
        int
            number of samples in the column with given name
        """
        recs = self._traverse_column_data_records(column_name, keys=True, values=False)
        return ilen(recs)  # regular len method not defined for generator iterable

# ------------------------- process schema ----------------------------------------------

    def schema_specs(self):
        """Return the all schema specs defined by all columns.

        Returns
        -------
        dict
            dict of column spec key and digest for each column schema
        """
        recs = {}
        for schema_key, schema_val in self._traverse_column_schema_records():
            schema_record = schema_column_record_from_db_key(schema_key)
            schema_val = data_record_digest_val_from_db_val(schema_val)
            recs[schema_record] = schema_val
        return recs

    def schema_hashes(self) -> List[str]:
        """Find all schema hashes inside of a commit

        Returns
        -------
        List[str]
            list of all schema hash digests in the commit
        """
        all_schema_hashes = []
        for schema_rec_val in self._traverse_column_schema_records(keys=False, values=True):
            digest = data_record_digest_val_from_db_val(schema_rec_val)
            all_schema_hashes.append(digest.digest)
        return all_schema_hashes

    def data_hash_to_schema_hash(self) -> Dict[str, str]:
        """For all hashes in the commit, map sample hash to schema hash.

        Returns
        -------
        Dict[str, str]
            mapping of sample hash to aset_schema_hash
        """
        odict = {}
        aset_names = self.column_names()
        aset_schema_specs = self.schema_specs()
        col_names_schema_digests = {k.column: v.digest for k, v in aset_schema_specs.items()}
        for asetn in aset_names:
            aset_hash_vals = self.column_data_hashes(asetn)
            aset_schema_hash = col_names_schema_digests[asetn]
            for aset_hash_val in aset_hash_vals:
                odict[aset_hash_val.digest] = aset_schema_hash
        return odict



from pathlib import Path
import warnings

import lmdb
from tqdm import tqdm

from ..records import (
    hash_data_db_key_from_raw_key,
    hash_schema_db_key_from_raw_key,
)
from ..backends import BACKEND_ACCESSOR_MAP
from ..txnctx import TxnRegister
from ..records import commiting, hashmachine, hashs, parsing, queries, heads
from ..op_state import report_corruption_risk_on_parsing_error


@report_corruption_risk_on_parsing_error
def _verify_column_integrity(hashenv: lmdb.Environment, repo_path: Path):

    hq = hashs.HashQuery(hashenv)
    narrays, nremote = hq.num_data_records(), 0
    array_kvs = hq.gen_all_data_digests_and_parsed_backend_specs()
    try:
        bes = {}
        for digest, spec in tqdm(array_kvs, total=narrays, desc='verifying column data'):
            if spec.backend not in bes:
                bes[spec.backend] = BACKEND_ACCESSOR_MAP[spec.backend](repo_path, None, None)
                bes[spec.backend].open(mode='r')
            if spec.islocal is False:
                nremote += 1
                continue
            data = bes[spec.backend].read_data(spec)
            tcode = hashmachine.hash_type_code_from_digest(digest)

            hash_func = hashmachine.hash_func_from_tcode(tcode)
            calc_digest = hash_func(data)
            if calc_digest != digest:
                raise RuntimeError(
                    f'Data corruption detected for array. Expected digest `{digest}` '
                    f'currently mapped to spec `{spec}`. Found digest `{calc_digest}`')
        if nremote > 0:
            warnings.warn(
                'Can not verify integrity of partially fetched array data references. '
                f'For complete proof, fetch all remote data locally. Did not verify '
                f'{nremote}/{narrays} arrays', RuntimeWarning)
    finally:
        for be in bes.keys():
            bes[be].close()


@report_corruption_risk_on_parsing_error
def _verify_schema_integrity(hashenv: lmdb.Environment):

    hq = hashs.HashQuery(hashenv)
    schema_kvs = hq.gen_all_schema_digests_and_parsed_specs()
    nschemas = hq.num_schema_records()
    for digest, val in tqdm(schema_kvs, total=nschemas, desc='verifying schemas'):
        tcode = hashmachine.hash_type_code_from_digest(digest)
        hash_func = hashmachine.hash_func_from_tcode(tcode)
        calc_digest = hash_func(val)
        if calc_digest != digest:
            raise RuntimeError(
                f'Data corruption detected for schema. Expected digest `{digest}` '
                f'currently mapped to spec `{val}`. Found digest `{calc_digest}`')


@report_corruption_risk_on_parsing_error
def _verify_commit_tree_integrity(refenv: lmdb.Environment):

    initialCmt = None
    all_commits = set(commiting.list_all_commits(refenv))
    reftxn = TxnRegister().begin_reader_txn(refenv)
    try:
        for cmt in tqdm(all_commits, desc='verifying commit trees'):
            pKey = parsing.commit_parent_db_key_from_raw_key(cmt)
            pVal = reftxn.get(pKey, default=False)
            if pVal is False:
                raise RuntimeError(
                    f'Data corruption detected for parent ref of commit `{cmt}`. '
                    f'Parent ref not recorded in refs db.')

            p_val = parsing.commit_parent_raw_val_from_db_val(pVal)
            parents = p_val.ancestor_spec
            if parents.master_ancestor != '':
                if parents.master_ancestor not in all_commits:
                    raise RuntimeError(
                        f'Data corruption detected in commit tree. Commit `{cmt}` '
                        f'with ancestors val `{parents}` references non-existing '
                        f'master ancestor `{parents.master_ancestor}`.')
            if parents.dev_ancestor != '':
                if parents.dev_ancestor not in all_commits:
                    raise RuntimeError(
                        f'Data corruption detected in commit tree. Commit `{cmt}` '
                        f'with ancestors val `{parents}` references non-existing '
                        f'dev ancestor `{parents.dev_ancestor}`.')
            if (parents.master_ancestor == '') and (parents.dev_ancestor == ''):
                if initialCmt is not None:
                    raise RuntimeError(
                        f'Commit tree integrity compromised. Multiple "initial" (commits '
                        f'with no parents) found. First `{initialCmt}`, second `{cmt}`')
                else:
                    initialCmt = cmt
    finally:
        TxnRegister().abort_reader_txn(refenv)


@report_corruption_risk_on_parsing_error
def _verify_commit_ref_digests_exist(hashenv: lmdb.Environment, refenv: lmdb.Environment):

    all_commits = commiting.list_all_commits(refenv)
    datatxn = TxnRegister().begin_reader_txn(hashenv, buffer=True)
    try:
        with datatxn.cursor() as cur:
            for cmt in tqdm(all_commits, desc='verifying commit ref digests'):
                with commiting.tmp_cmt_env(refenv, cmt) as tmpDB:
                    rq = queries.RecordQuery(tmpDB)
                    array_data_digests = set(rq.data_hashes())
                    schema_digests = set(rq.schema_hashes())

                    for datadigest in array_data_digests:
                        dbk = hash_data_db_key_from_raw_key(datadigest)
                        exists = cur.set_key(dbk)
                        if exists is False:
                            raise RuntimeError(
                                f'Data corruption detected in commit refs. Commit `{cmt}` '
                                f'references array data digest `{datadigest}` which does not '
                                f'exist in data hash db.')

                    for schemadigest in schema_digests:
                        dbk = hash_schema_db_key_from_raw_key(schemadigest)
                        exists = cur.set_key(dbk)
                        if exists is False:
                            raise RuntimeError(
                                f'Data corruption detected in commit refs. Commit `{cmt}` '
                                f'references schema digest `{schemadigest}` which does not '
                                f'exist in data hash db.')

    finally:
        TxnRegister().abort_reader_txn(hashenv)


@report_corruption_risk_on_parsing_error
def _verify_branch_integrity(branchenv: lmdb.Environment, refenv: lmdb.Environment):

    branch_names = heads.get_branch_names(branchenv)
    if len(branch_names) < 1:
        raise RuntimeError(
            f'Branch map compromised. Repo must contain atleast one branch. '
            f'Found {len(branch_names)} branches.')

    for bname in tqdm(branch_names, desc='verifying branches'):
        bhead = heads.get_branch_head_commit(branchenv=branchenv, branch_name=bname)
        exists = commiting.check_commit_hash_in_history(refenv=refenv, commit_hash=bhead)
        if exists is False:
            raise RuntimeError(
                f'Branch commit map compromised. Branch name `{bname}` references '
                f'commit digest `{bhead}` which does not exist in refs db.')

    staging_bname = heads.get_staging_branch_head(branchenv)
    if staging_bname not in branch_names:
        raise RuntimeError(
            f'Brach commit map compromised. Staging head refers to branch name '
            f'`{staging_bname}` which does not exist in the branch db.')


def run_verification(branchenv: lmdb.Environment,
                     hashenv: lmdb.Environment,
                     refenv: lmdb.Environment,
                     repo_path: Path):

    _verify_branch_integrity(branchenv, refenv)
    _verify_commit_tree_integrity(refenv)
    _verify_commit_ref_digests_exist(hashenv, refenv)
    _verify_schema_integrity(hashenv)
    _verify_column_integrity(hashenv, repo_path)



"""Local Numpy memmap Backend Implementation, Identifier: ``NUMPY_10``

Backend Identifiers
===================

*  Backend: ``1``
*  Version: ``0``
*  Format Code: ``10``
*  Canonical Name: ``NUMPY_10``

Storage Method
==============

* Data is written to specific subarray indexes inside a numpy memmapped array on disk.

* Each file is a zero-initialized array of

  *  ``dtype: {schema_dtype}``; ie ``np.float32`` or ``np.uint8``

  *  ``shape: (COLLECTION_SIZE, *{schema_shape})``; ie ``(500, 10)`` or ``(500,
     4, 3)``. The first index in the array is referred to as a "collection
     index".

Compression Options
===================

Does not accept any compression options. No compression is applied.

Record Format
=============

Fields Recorded for Each Array
------------------------------

*  Format Code
*  File UID
*  xxhash64_hexdigest
*  Collection Index (0:COLLECTION_SIZE subarray selection)
*  Subarray Shape

Examples
--------

1)  Adding the first piece of data to a file:

    *  Array shape (Subarray Shape): (10, 10)
    *  File UID: "K3ktxv"
    *  xxhash64_hexdigest: 94701dd9f32626e2
    *  Collection Index: 488

    ``Record Data =>  "10:K3ktxv:94701dd9f32626e2:488:10 10"``

2)  Adding to a piece of data to a the middle of a file:

    *  Array shape (Subarray Shape): (20, 2, 3)
    *  File UID: "Mk23nl"
    *  xxhash64_hexdigest: 1363344b6c051b29
    *  Collection Index: 199

    ``Record Data => "10:Mk23nl:1363344b6c051b29:199:20 2 3"``


Technical Notes
===============

*  A typical numpy memmap file persisted to disk does not retain information
   about its datatype or shape, and as such must be provided when re-opened
   after close. In order to persist a memmap in ``.npy`` format, we use the a
   special function ``open_memmap`` imported from ``np.lib.format`` which can
   open a memmap file and persist necessary header info to disk in ``.npy``
   format.

*  On each write, an ``xxhash64_hexdigest`` checksum is calculated. This is not
   for use as the primary hash algorithm, but rather stored in the local record
   format itself to serve as a quick way to verify no disk corruption occurred.
   This is required since numpy has no built in data integrity validation
   methods when reading from disk.
"""
import os
from collections import ChainMap
from functools import partial
from pathlib import Path
from typing import MutableMapping, Optional

import numpy as np
from numpy.lib.format import open_memmap
from xxhash import xxh64_hexdigest

from .specs import NUMPY_10_DataHashSpec
from ..constants import DIR_DATA_REMOTE, DIR_DATA_STAGE, DIR_DATA_STORE, DIR_DATA
from ..op_state import reader_checkout_only, writer_checkout_only
from ..utils import random_string
from ..typesystem import Descriptor, OneOf, EmptyDict, checkedmeta


# ----------------------------- Configuration ---------------------------------

_FmtCode = '10'

# number of subarray contents of a single numpy memmap file
COLLECTION_SIZE = 1000

# -------------------------------- Parser Implementation ----------------------


def numpy_10_encode(uid: str, cksum: str, collection_idx: int, shape: tuple) -> bytes:
    """converts the numpy data spect to an appropriate db value

    Parameters
    ----------
    uid : str
        file name (schema uid) of the np file to find this data piece in.
    cksum : int
        xxhash64_hexdigest checksum of the data as computed on that local machine.
    collection_idx : int
        collection first axis index in which this data piece resides.
    shape : tuple
        shape of the data sample written to the collection idx. ie: what
        subslices of the array should be read to retrieve the sample as
        recorded.

    Returns
    -------
    bytes
        hash data db value recording all input specifications
    """
    shape_str = " ".join([str(i) for i in shape])
    return f'10:{uid}:{cksum}:{collection_idx}:{shape_str}'.encode()


# ------------------------- Accessor Object -----------------------------------


@OneOf(list(map(lambda x: np.dtype(x).name, [
        np.bool, np.uint8, np.uint16, np.uint32, np.uint64, np.int8, np.int16,
        np.int32, np.int64, np.float16, np.float32, np.float64, np.longdouble])))
class AllowedDtypes(Descriptor):
    # Note. np.longdouble since np.float128 not guaranteed to be available on
    # all system. this is a particular issue with some windows numpy builds.
    pass


class NUMPY_10_Options(metaclass=checkedmeta):
    _dtype = AllowedDtypes()
    _backend_options = EmptyDict()

    def __init__(self, backend_options, dtype, *args, **kwargs):
        if backend_options is None:
            backend_options = self.default_options

        self._backend_options = backend_options
        self._dtype = dtype

    @property
    def default_options(self):
        return {}

    @property
    def backend_options(self):
        return self._backend_options

    @property
    def init_requires(self):
        return ('repo_path', 'schema_shape', 'schema_dtype')


class NUMPY_10_FileHandles(object):

    def __init__(self, repo_path: Path, schema_shape: tuple, schema_dtype: np.dtype):
        self.repo_path = repo_path
        self.schema_shape = schema_shape
        self.schema_dtype = schema_dtype
        self._dflt_backend_opts: Optional[dict] = None

        self.rFp: MutableMapping[str, np.memmap] = {}
        self.wFp: MutableMapping[str, np.memmap] = {}
        self.Fp = ChainMap(self.rFp, self.wFp)

        self.mode: str = None
        self.w_uid: str = None
        self.hIdx: int = None

        self.STAGEDIR: Path = Path(self.repo_path, DIR_DATA_STAGE, _FmtCode)
        self.REMOTEDIR: Path = Path(self.repo_path, DIR_DATA_REMOTE, _FmtCode)
        self.DATADIR: Path = Path(self.repo_path, DIR_DATA, _FmtCode)
        self.STOREDIR: Path = Path(self.repo_path, DIR_DATA_STORE, _FmtCode)
        self.DATADIR.mkdir(exist_ok=True)

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        self.close()
        state = self.__dict__.copy()
        del state['rFp']
        del state['wFp']
        del state['Fp']
        return state

    def __setstate__(self, state: dict) -> None:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.__dict__.update(state)
        self.rFp = {}
        self.wFp = {}
        self.Fp = ChainMap(self.rFp, self.wFp)
        self.open(mode=self.mode)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        if self.w_uid in self.wFp:
            self.wFp[self.w_uid].flush()

    @property
    def backend_opts(self):
        return self._dflt_backend_opts

    @writer_checkout_only
    def _backend_opts_set(self, val):
        """Nonstandard descriptor method. See notes in ``backend_opts.setter``.
        """
        self._dflt_backend_opts = val
        return

    @backend_opts.setter
    def backend_opts(self, value):
        """
        Using seperate setter method (with ``@writer_checkout_only`` decorator
        applied) due to bug in python <3.8.

        From: https://bugs.python.org/issue19072
            > The classmethod decorator when applied to a function of a class,
            > does not honour the descriptor binding protocol for whatever it
            > wraps. This means it will fail when applied around a function which
            > has a decorator already applied to it and where that decorator
            > expects that the descriptor binding protocol is executed in order
            > to properly bind the function to the class.
        """
        return self._backend_opts_set(value)

    def open(self, mode: str, *, remote_operation: bool = False):
        """open numpy file handle coded directories

        Parameters
        ----------
        mode : str
            one of `a` for `write-enabled` mode or `r` for read-only
        remote_operation : bool, optional, kwarg only
            True if remote operations call this method. Changes the symlink
            directories used while writing., by default False
        """
        self.mode = mode
        if self.mode == 'a':
            process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
            process_dir.mkdir(exist_ok=True)
            for uidpth in process_dir.iterdir():
                if uidpth.suffix == '.npy':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(open_memmap, file_pth, 'r')

        if not remote_operation:
            if not self.STOREDIR.is_dir():
                return
            for uidpth in self.STOREDIR.iterdir():
                if uidpth.suffix == '.npy':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(open_memmap, file_pth, 'r')

    def close(self, *args, **kwargs):
        """Close any open file handles.
        """
        if self.mode == 'a':
            if self.w_uid in self.wFp:
                self.wFp[self.w_uid].flush()
                self.w_uid = None
                self.hIdx = None
            for k in list(self.wFp.keys()):
                del self.wFp[k]

        for k in list(self.rFp.keys()):
            del self.rFp[k]

    @staticmethod
    def delete_in_process_data(repo_path: Path, *, remote_operation: bool = False):
        """Removes some set of files entirely from the stage/remote directory.

        DANGER ZONE. This should essentially only be used to perform hard resets
        of the repository state.

        Parameters
        ----------
        repo_path : Path
            path to the repository on disk
        remote_operation : optional, kwarg only, bool
            If true, modify contents of the remote_dir, if false (default) modify
            contents of the staging directory.
        """
        data_dir = Path(repo_path, DIR_DATA, _FmtCode)
        pdir = DIR_DATA_STAGE if not remote_operation else DIR_DATA_REMOTE
        process_dir = Path(repo_path, pdir, _FmtCode)
        if not process_dir.is_dir():
            return

        for uidpth in process_dir.iterdir():
            if uidpth.suffix == '.npy':
                os.remove(process_dir.joinpath(uidpth.name))
                os.remove(data_dir.joinpath(uidpth.name))
        os.rmdir(process_dir)

    def _create_schema(self, *, remote_operation: bool = False):
        """stores the shape and dtype as the schema of a column.

        Parameters
        ----------
        remote_operation : optional, kwarg only, bool
            if this schema is being created from a remote fetch operation, then do not
            place the file symlink in the staging directory. Instead symlink it
            to a special remote staging directory. (default is False, which places the
            symlink in the stage data directory.)
        """
        uid = random_string()
        file_path = self.DATADIR.joinpath(f'{uid}.npy')
        m = open_memmap(file_path,
                        mode='w+',
                        dtype=self.schema_dtype,
                        shape=(COLLECTION_SIZE, *self.schema_shape))
        self.wFp[uid] = m
        self.w_uid = uid
        self.hIdx = 0

        process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
        Path(process_dir, f'{uid}.npy').touch()

    def read_data(self, hashVal: NUMPY_10_DataHashSpec) -> np.ndarray:
        """Read data from disk written in the numpy_00 fmtBackend

        Parameters
        ----------
        hashVal : NUMPY_10_DataHashSpec
            record specification stored in the db

        Returns
        -------
        np.ndarray
            tensor data stored at the provided hashVal specification.

        Raises
        ------
        RuntimeError
            If the recorded checksum does not match the received checksum.

        Notes
        -----

        TO AVOID DATA LOSS / CORRUPTION:

        * On a read operation, we copy memmap subarray tensor data to a new
          `np.ndarray` instance so as to prevent writes on a raw memmap result
          slice (a `np.memmap` instance) from propogating to data on disk.

        * This is an issue for reads from a write-enabled checkout where data
          was just written, since the np flag "WRITEABLE" and "OWNDATA" will be
          true, and writes to the returned array would be overwrite that data
          slice on disk.

        * For read-only checkouts, modifications to the resultant array would
          perform a "copy on write"-like operation which would be propogated to
          all future reads of the subarray from that process, but which would
          not be persisted to disk.
        """
        srcSlc = (hashVal.collection_idx, *[slice(0, x) for x in hashVal.shape])
        try:
            res = self.Fp[hashVal.uid][srcSlc]
        except TypeError:
            self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
            res = self.Fp[hashVal.uid][srcSlc]
        except KeyError:
            process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
            if Path(process_dir, f'{hashVal.uid}.npy').is_file():
                file_pth = self.DATADIR.joinpath(f'{hashVal.uid}.npy')
                self.rFp[hashVal.uid] = open_memmap(file_pth, 'r')
                res = self.Fp[hashVal.uid][srcSlc]
            else:
                raise

        out = np.array(res, dtype=res.dtype, order='C')
        if xxh64_hexdigest(out) != hashVal.checksum:
            raise RuntimeError(
                f'DATA CORRUPTION Checksum {xxh64_hexdigest(out)} != recorded {hashVal}')
        return out

    def write_data(self, array: np.ndarray, *, remote_operation: bool = False) -> bytes:
        """writes array data to disk in the numpy_00 fmtBackend

        Parameters
        ----------
        array : np.ndarray
            tensor to write to disk
        remote_operation : bool, optional, kwarg only
            True if writing in a remote operation, otherwise False. Default is
            False

        Returns
        -------
        bytes
            db hash record value specifying location information
        """
        checksum = xxh64_hexdigest(array)
        if self.w_uid in self.wFp:
            self.hIdx += 1
            if self.hIdx >= COLLECTION_SIZE:
                self.wFp[self.w_uid].flush()
                self._create_schema(remote_operation=remote_operation)
        else:
            self._create_schema(remote_operation=remote_operation)

        destSlc = (self.hIdx, *[slice(0, x) for x in array.shape])
        self.wFp[self.w_uid][destSlc] = array
        self.wFp[self.w_uid].flush()
        return numpy_10_encode(self.w_uid, checksum, self.hIdx, array.shape)



from .base import ColumnBase
from .descriptors import OneOf, Descriptor, String, OptionalString, OptionalDict
from ..records import CompatibleData
from ..utils import format_bytes


@OneOf(['<class\'bytes\'>'])
class BytesDType(Descriptor):
    pass


SERIAL_DTYPE_TO_OBJ = {
    '<class\'bytes\'>': bytes,
}


@OneOf(['variable_shape'])
class BytesSchemaType(String):
    pass


@OneOf(['bytes'])
class BytesColumnType(String):
    pass


@OneOf(['3'])
class DataHasherTcode(String):
    pass


class BytesSchemaBase(ColumnBase):
    _schema_type = BytesSchemaType()
    _column_type = BytesColumnType()
    _data_hasher_tcode = DataHasherTcode()

    def __init__(
            self,
            dtype,
            backend=None,
            backend_options=None,
            *args, **kwargs
    ):
        if 'data_hasher_tcode' not in kwargs:
            kwargs['data_hasher_tcode'] = '3'
        super().__init__(*args, **kwargs)

        if backend_options is not None and backend is None:
            raise ValueError(
                '`backend_options` cannot be set if `backend` is not also provided.')

        if not isinstance(dtype, str):
            dtype = repr(dtype).replace(' ', '')

        self._dtype = dtype
        self._backend = backend
        self._backend_options = backend_options
        self._schema_attributes.extend(
            ['_schema_type', '_dtype', '_backend', '_backend_options']
        )

    def backend_from_heuristics(self):
        self._backend = '31'

    @property
    def schema_type(self):
        return self._schema_type

    @property
    def dtype(self):
        return SERIAL_DTYPE_TO_OBJ[self._dtype]

    @property
    def backend(self):
        return self._backend

    @property
    def backend_options(self):
        return self._backend_options

    def data_hash_digest(self, data: str) -> str:
        return self._data_hasher_func(data)

    def change_backend(self, backend, backend_options=None):
        old_backend = self._backend
        old_backend_options = self._backend_options
        try:
            self._backend = backend
            self._backend_options = backend_options
            # del and reset beopts object to reverify input correctness.
            del self._beopts
            self._backend_options = self._beopts.backend_options
        except (TypeError, ValueError) as e:
            del self._beopts
            self._backend = old_backend
            self._backend_options = old_backend_options
            self._backend_options = self._beopts.backend_options
            raise e from None


@OneOf(['31', '50', None])
class BytesVariableShapeBackends(OptionalString):
    pass


@OneOf(['variable_shape'])
class VariableShapeSchemaType(String):
    pass


class BytesVariableShape(BytesSchemaBase):
    _dtype = BytesDType()
    _backend = BytesVariableShapeBackends()
    _backend_options = OptionalDict()
    _schema_type = VariableShapeSchemaType()

    def __init__(self, *args, **kwargs):
        if 'column_type' in kwargs:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(column_type='bytes', *args, **kwargs)

        if 'schema_type' in kwargs:
            self._schema_type = kwargs['schema_type']
        else:
            self._schema_type = 'variable_shape'

        if self.backend is None:
            self.backend_from_heuristics()
        self._backend_options = self._beopts.backend_options

    def verify_data_compatible(self, data):
        compatible = True
        reason = ''
        if not isinstance(data, bytes):
            compatible = False
            reason = f'data {data} not valid, must be of type {bytes} not{type(data)}'
        elif len(data) > 2000000:  # 2MB
            compatible = False
            reason = f'bytes must be less than 2MB in size, recieved {format_bytes(len(data))}'

        res = CompatibleData(compatible, reason)
        return res



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import hangar_service_pb2 as hangar__service__pb2


class HangarServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.PING = channel.unary_unary(
        '/hangar.HangarService/PING',
        request_serializer=hangar__service__pb2.PingRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.PingReply.FromString,
        )
    self.GetClientConfig = channel.unary_unary(
        '/hangar.HangarService/GetClientConfig',
        request_serializer=hangar__service__pb2.GetClientConfigRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.GetClientConfigReply.FromString,
        )
    self.FetchBranchRecord = channel.unary_unary(
        '/hangar.HangarService/FetchBranchRecord',
        request_serializer=hangar__service__pb2.FetchBranchRecordRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FetchBranchRecordReply.FromString,
        )
    self.FetchData = channel.stream_stream(
        '/hangar.HangarService/FetchData',
        request_serializer=hangar__service__pb2.FetchDataRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FetchDataReply.FromString,
        )
    self.FetchCommit = channel.unary_stream(
        '/hangar.HangarService/FetchCommit',
        request_serializer=hangar__service__pb2.FetchCommitRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FetchCommitReply.FromString,
        )
    self.FetchSchema = channel.unary_unary(
        '/hangar.HangarService/FetchSchema',
        request_serializer=hangar__service__pb2.FetchSchemaRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FetchSchemaReply.FromString,
        )
    self.PushBranchRecord = channel.unary_unary(
        '/hangar.HangarService/PushBranchRecord',
        request_serializer=hangar__service__pb2.PushBranchRecordRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.PushBranchRecordReply.FromString,
        )
    self.PushData = channel.stream_unary(
        '/hangar.HangarService/PushData',
        request_serializer=hangar__service__pb2.PushDataRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.PushDataReply.FromString,
        )
    self.PushCommit = channel.stream_unary(
        '/hangar.HangarService/PushCommit',
        request_serializer=hangar__service__pb2.PushCommitRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.PushCommitReply.FromString,
        )
    self.PushSchema = channel.unary_unary(
        '/hangar.HangarService/PushSchema',
        request_serializer=hangar__service__pb2.PushSchemaRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.PushSchemaReply.FromString,
        )
    self.FetchFindMissingCommits = channel.unary_unary(
        '/hangar.HangarService/FetchFindMissingCommits',
        request_serializer=hangar__service__pb2.FindMissingCommitsRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingCommitsReply.FromString,
        )
    self.FetchFindMissingHashRecords = channel.stream_stream(
        '/hangar.HangarService/FetchFindMissingHashRecords',
        request_serializer=hangar__service__pb2.FindMissingHashRecordsRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingHashRecordsReply.FromString,
        )
    self.FetchFindMissingSchemas = channel.unary_unary(
        '/hangar.HangarService/FetchFindMissingSchemas',
        request_serializer=hangar__service__pb2.FindMissingSchemasRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingSchemasReply.FromString,
        )
    self.PushFindMissingCommits = channel.unary_unary(
        '/hangar.HangarService/PushFindMissingCommits',
        request_serializer=hangar__service__pb2.FindMissingCommitsRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingCommitsReply.FromString,
        )
    self.PushFindMissingHashRecords = channel.stream_stream(
        '/hangar.HangarService/PushFindMissingHashRecords',
        request_serializer=hangar__service__pb2.FindMissingHashRecordsRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingHashRecordsReply.FromString,
        )
    self.PushFindMissingSchemas = channel.unary_unary(
        '/hangar.HangarService/PushFindMissingSchemas',
        request_serializer=hangar__service__pb2.FindMissingSchemasRequest.SerializeToString,
        response_deserializer=hangar__service__pb2.FindMissingSchemasReply.FromString,
        )


class HangarServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def PING(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetClientConfig(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchBranchRecord(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchData(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchCommit(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchSchema(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushBranchRecord(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushData(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushCommit(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushSchema(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchFindMissingCommits(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchFindMissingHashRecords(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FetchFindMissingSchemas(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushFindMissingCommits(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushFindMissingHashRecords(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushFindMissingSchemas(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_HangarServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'PING': grpc.unary_unary_rpc_method_handler(
          servicer.PING,
          request_deserializer=hangar__service__pb2.PingRequest.FromString,
          response_serializer=hangar__service__pb2.PingReply.SerializeToString,
      ),
      'GetClientConfig': grpc.unary_unary_rpc_method_handler(
          servicer.GetClientConfig,
          request_deserializer=hangar__service__pb2.GetClientConfigRequest.FromString,
          response_serializer=hangar__service__pb2.GetClientConfigReply.SerializeToString,
      ),
      'FetchBranchRecord': grpc.unary_unary_rpc_method_handler(
          servicer.FetchBranchRecord,
          request_deserializer=hangar__service__pb2.FetchBranchRecordRequest.FromString,
          response_serializer=hangar__service__pb2.FetchBranchRecordReply.SerializeToString,
      ),
      'FetchData': grpc.stream_stream_rpc_method_handler(
          servicer.FetchData,
          request_deserializer=hangar__service__pb2.FetchDataRequest.FromString,
          response_serializer=hangar__service__pb2.FetchDataReply.SerializeToString,
      ),
      'FetchCommit': grpc.unary_stream_rpc_method_handler(
          servicer.FetchCommit,
          request_deserializer=hangar__service__pb2.FetchCommitRequest.FromString,
          response_serializer=hangar__service__pb2.FetchCommitReply.SerializeToString,
      ),
      'FetchSchema': grpc.unary_unary_rpc_method_handler(
          servicer.FetchSchema,
          request_deserializer=hangar__service__pb2.FetchSchemaRequest.FromString,
          response_serializer=hangar__service__pb2.FetchSchemaReply.SerializeToString,
      ),
      'PushBranchRecord': grpc.unary_unary_rpc_method_handler(
          servicer.PushBranchRecord,
          request_deserializer=hangar__service__pb2.PushBranchRecordRequest.FromString,
          response_serializer=hangar__service__pb2.PushBranchRecordReply.SerializeToString,
      ),
      'PushData': grpc.stream_unary_rpc_method_handler(
          servicer.PushData,
          request_deserializer=hangar__service__pb2.PushDataRequest.FromString,
          response_serializer=hangar__service__pb2.PushDataReply.SerializeToString,
      ),
      'PushCommit': grpc.stream_unary_rpc_method_handler(
          servicer.PushCommit,
          request_deserializer=hangar__service__pb2.PushCommitRequest.FromString,
          response_serializer=hangar__service__pb2.PushCommitReply.SerializeToString,
      ),
      'PushSchema': grpc.unary_unary_rpc_method_handler(
          servicer.PushSchema,
          request_deserializer=hangar__service__pb2.PushSchemaRequest.FromString,
          response_serializer=hangar__service__pb2.PushSchemaReply.SerializeToString,
      ),
      'FetchFindMissingCommits': grpc.unary_unary_rpc_method_handler(
          servicer.FetchFindMissingCommits,
          request_deserializer=hangar__service__pb2.FindMissingCommitsRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingCommitsReply.SerializeToString,
      ),
      'FetchFindMissingHashRecords': grpc.stream_stream_rpc_method_handler(
          servicer.FetchFindMissingHashRecords,
          request_deserializer=hangar__service__pb2.FindMissingHashRecordsRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingHashRecordsReply.SerializeToString,
      ),
      'FetchFindMissingSchemas': grpc.unary_unary_rpc_method_handler(
          servicer.FetchFindMissingSchemas,
          request_deserializer=hangar__service__pb2.FindMissingSchemasRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingSchemasReply.SerializeToString,
      ),
      'PushFindMissingCommits': grpc.unary_unary_rpc_method_handler(
          servicer.PushFindMissingCommits,
          request_deserializer=hangar__service__pb2.FindMissingCommitsRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingCommitsReply.SerializeToString,
      ),
      'PushFindMissingHashRecords': grpc.stream_stream_rpc_method_handler(
          servicer.PushFindMissingHashRecords,
          request_deserializer=hangar__service__pb2.FindMissingHashRecordsRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingHashRecordsReply.SerializeToString,
      ),
      'PushFindMissingSchemas': grpc.unary_unary_rpc_method_handler(
          servicer.PushFindMissingSchemas,
          request_deserializer=hangar__service__pb2.FindMissingSchemasRequest.FromString,
          response_serializer=hangar__service__pb2.FindMissingSchemasReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'hangar.HangarService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from pathlib import Path
from typing import Iterable, List, Tuple, Union

import lmdb

from .column_parsers import (
    hash_record_count_start_range_key,
    hash_schema_raw_key_from_db_key,
    hash_data_raw_key_from_db_key,
    schema_hash_db_key_from_digest,
    schema_spec_from_db_val,
    schema_record_count_start_range_key
)
from ..backends import BACKEND_ACCESSOR_MAP, backend_decoder
from ..txnctx import TxnRegister
from ..mixins import CursorRangeIterator
from ..utils import ilen


class HashQuery(CursorRangeIterator):
    """Traverse and query contents contained in ``hashenv`` db

    These methods operate on the database which store the mapping of some data
    digest to it's location on disk (or value in the case of metadata and
    schemas). These databases are not specific to a particular commit; the
    records are for every piece of data stored in every commit across history.

    There are relatively few procedures which require traversal and mapping
    across data records in this manner. The two most notable use cases are:

        1. Remote client-server negotiation operations
        2. Verifying the integrity of a repositories historical provenance, commit
        contents, and data stored on disk.
    """

    def __init__(self, hashenv: lmdb.Environment):
        self._hashenv = hashenv

    # ------------------ traversing the unpacked records ----------------------

    def _traverse_all_hash_records(self, keys: bool = True, values: bool = True
                                   ) -> Iterable[Union[bytes, Tuple[bytes, bytes]]]:
        """PUll out all binary encoded data hash records.

        Parameters
        ----------
        keys : bool, optional
            if True, returns keys, by default True
        values : bool, optional
            if True, return values, by default True

        Yields
        -------
        Union[bytes, Tuple[bytes, bytes]]
            Iterable of schema record keys, values, or items tuple
        """
        startHashRangeKey = hash_record_count_start_range_key()
        try:
            hashtxn = TxnRegister().begin_reader_txn(self._hashenv)
            yield from self.cursor_range_iterator(hashtxn, startHashRangeKey, keys, values)
        finally:
            TxnRegister().abort_reader_txn(self._hashenv)

    def _traverse_all_schema_records(self, keys: bool = True, values: bool = True
                                     ) -> Iterable[Union[bytes, Tuple[bytes, bytes]]]:
        """Pull out all binary encoded schema hash records.

        Parameters
        ----------
        keys : bool, optional
            if True, returns keys, by default True
        values : bool, optional
            if True, return values, by default True

        Yields
        -------
        Union[bytes, Tuple[bytes, bytes]]
            Iterable of schema record keys, values, or items tuple
        """
        startSchemaRangeKey = schema_record_count_start_range_key()
        try:
            hashtxn = TxnRegister().begin_reader_txn(self._hashenv)
            yield from self.cursor_range_iterator(hashtxn, startSchemaRangeKey, keys, values)
        finally:
            TxnRegister().abort_reader_txn(self._hashenv)

    def list_all_hash_keys_raw(self) -> List[str]:
        recs = self._traverse_all_hash_records(keys=True, values=False)
        return list(map(hash_data_raw_key_from_db_key, recs))

    def gen_all_hash_keys_db(self) -> Iterable[bytes]:
        return self._traverse_all_hash_records(keys=True, values=False)

    def list_all_schema_digests(self) -> List[str]:
        recs = self._traverse_all_schema_records(keys=True, values=False)
        return list(map(hash_schema_raw_key_from_db_key, recs))

    def gen_all_schema_keys_db(self) -> Iterable[bytes]:
        return self._traverse_all_schema_records(keys=True, values=False)

    def num_data_records(self) -> int:
        """Total count of all data digests / backends specs stored over all repo history.
        """
        num_total = self._hashenv.stat()['entries']
        remaining = num_total - self.num_schema_records()
        return remaining

    def num_schema_records(self) -> int:
        """Total count of schema digests / spec defs stored over all repo history.
        """
        return ilen(self._traverse_all_schema_records(keys=True, values=False))

    def gen_all_data_digests_and_parsed_backend_specs(self):
        for dbk, dbv in self._traverse_all_hash_records(keys=True, values=True):
            rawk = hash_data_raw_key_from_db_key(dbk)
            rawv = backend_decoder(dbv)
            yield (rawk, rawv)

    def gen_all_schema_digests_and_parsed_specs(self) -> Iterable[Tuple[str, dict]]:
        for dbk, dbv in self._traverse_all_schema_records(keys=True, values=True):
            rawk = hash_schema_raw_key_from_db_key(dbk)
            rawv = schema_spec_from_db_val(dbv)
            yield (rawk, rawv)

    def get_schema_digest_spec(self, digest) -> dict:
        schemaHashKey = schema_hash_db_key_from_digest(digest)
        try:
            hashtxn = TxnRegister().begin_reader_txn(self._hashenv)
            schemaSpecVal = hashtxn.get(schemaHashKey)
        finally:
            TxnRegister().abort_reader_txn(self._hashenv)

        schema_spec = schema_spec_from_db_val(schemaSpecVal)
        return schema_spec


def backends_remove_in_process_data(repo_path: Path, *, remote_operation: bool = False):
    """DANGER! Permanently delete uncommitted data files/links for stage or remote area.

    This searches each backend accessors staged (or remote) folder structure for
    files, and if any are present the symlinks in stagedir and backing data
    files in datadir are removed.

    Parameters
    ----------
    repo_path : Path
        path to the repository on disk
    remote_operation : optional, kwarg only, bool
        If true, modify contents of the remote_dir, if false (default) modify
        contents of the staging directory.
    """
    for backend, accesor in BACKEND_ACCESSOR_MAP.items():
        if accesor is not None:
            accesor.delete_in_process_data(repo_path=repo_path,
                                           remote_operation=remote_operation)


def clear_stage_hash_records(stagehashenv):
    """Drop all records in the stagehashenv db

    This operation should be performed anytime a reset of the staging area is
    performed (including for commits, merges, and checkouts)

    Parameters
    ----------
    stagehashenv : lmdb.Environment
        db where staged data hash additions are recorded
    """
    stagehashtxn = TxnRegister().begin_writer_txn(stagehashenv)
    with stagehashtxn.cursor() as cursor:
        positionExists = cursor.first()
        while positionExists:
            positionExists = cursor.delete()
    cursor.close()
    TxnRegister().commit_writer_txn(stagehashenv)


def remove_stage_hash_records_from_hashenv(hashenv, stagehashenv):
    """Remove references to data additions during a hard reset

    For every hash record in stagehashenv, remove the corresponding k/v pair
    from the hashenv db. This is a dangerous operation if the stagehashenv was
    not appropriately constructed!!!

    Parameters
    ----------
    hashenv : lmdb.Environment
        db where all the permanent hash records are stored
    stagehashenv : lmdb.Environment
        db where all the staged hash records to be removed are stored.
    """
    stageHashKeys = HashQuery(stagehashenv).gen_all_hash_keys_db()
    hashtxn = TxnRegister().begin_writer_txn(hashenv)
    for hashKey in stageHashKeys:
        hashtxn.delete(hashKey)
    TxnRegister().commit_writer_txn(hashenv)



from typing import NamedTuple, Union, Sequence, Tuple, List, Optional

import numpy as np

from ..columns.constructors import open_file_handles, column_type_object_from_schema
from ..context import Environments
from ..records import parsing
from ..records import (
    schema_spec_from_db_val,
    hash_schema_db_key_from_raw_key,
    hash_data_db_key_from_raw_key,
)
from ..txnctx import TxnRegister


class ContentWriter(object):
    """Common methods to client & server which write content received.

    These are special methods configured especially for remote operations.
    They do not honor the public facing API or data write/read conventions
    established for users or the rest of Hangar internals.

    Parameters
    ----------
    envs : context.Environments
        main hangar environment context object.
    """

    def __init__(self, envs):

        self.env: Environments = envs
        self.txnctx: TxnRegister = TxnRegister()

    def commit(self, commit: str, parentVal: bytes, specVal: bytes,
               refVal: bytes) -> Union[str, bool]:
        """Write a commit record to the ref db

        Parameters
        ----------
        commit : str
            commit hash to write
        parentVal : bytes
            db formatted representation of commit parents
        specVal : bytes
            db formatted representation of the commit specs
        refVal : bytes
            db formated representation of commit record contents

        Returns
        -------
        str or False
            Commit hash if operation was successful.

            False if the commit hash existed in the db previously and
            no records were written.
        """
        commitSpecKey = parsing.commit_spec_db_key_from_raw_key(commit)
        commitParentKey = parsing.commit_parent_db_key_from_raw_key(commit)
        commitRefKey = parsing.commit_ref_db_key_from_raw_key(commit)
        refTxn = self.txnctx.begin_writer_txn(self.env.refenv)
        try:
            cmtParExists = refTxn.put(commitParentKey, parentVal, overwrite=False)
            cmtRefExists = refTxn.put(commitRefKey, refVal, overwrite=False)
            cmtSpcExists = refTxn.put(commitSpecKey, specVal, overwrite=False)
        finally:
            self.txnctx.commit_writer_txn(self.env.refenv)

        ret = False if not all([cmtParExists, cmtRefExists, cmtSpcExists]) else commit
        return ret

    def schema(self, schema_hash: str, schemaVal: bytes) -> Union[str, bool]:
        """Write a column schema hash specification record to the db

        Parameters
        ----------
        schema_hash : str
            schema hash being written
        schemaVal : bytes
            db formatted representation of schema specification

        Returns
        -------
        str or False
            schema_hash written if operation was successful.

            False if the schema_hash existed in db and no records written.
        """
        schemaKey = hash_schema_db_key_from_raw_key(schema_hash)
        hashTxn = self.txnctx.begin_writer_txn(self.env.hashenv)
        try:
            schemaExists = hashTxn.put(schemaKey, schemaVal, overwrite=False)
        finally:
            self.txnctx.commit_writer_txn(self.env.hashenv)

        ret = False if not schemaExists else schema_hash
        return ret

    def data(self,
             schema_hash: str,
             received_data: Sequence[Tuple[str, np.ndarray]],
             backend: Optional[str] = None,
             backend_options: Optional[dict] = None) -> List[str]:
        """Write data content to the hash records database

        Parameters
        ----------
        schema_hash : str
            schema_hash currently being written
        received_data : Sequence[Tuple[str, np.ndarray]]
            list of tuples, each specifying (digest, tensor) for data retrieved
            from the server. However, if a backend is manually specified which
            requires different input to the ``write_data`` method than a tensor,
            the second element can be replaced with what is appropriate for that
            situation.
        backend : str, optional
            Manually specified backend code which will be used to record the
            data records. If not specified (``None``), the default backend
            recorded in the schema spec will be used, by default None

        Returns
        -------
        List[str]
            list of str of all data digests written by this method.
        """
        schemaKey = hash_schema_db_key_from_raw_key(schema_hash)
        hashTxn = self.txnctx.begin_reader_txn(self.env.hashenv)
        try:
            schemaVal = hashTxn.get(schemaKey)
        finally:
            self.txnctx.abort_reader_txn(self.env.hashenv)
        schema_val = schema_spec_from_db_val(schemaVal)
        schema = column_type_object_from_schema(schema_val)

        if (backend is not None) and ((backend != schema.backend) or (backend_options is not None)):
            schema.change_backend(backend, backend_options=backend_options)

        be_accessor = open_file_handles(backends=[schema.backend],
                                        path=self.env.repo_path,
                                        mode='a',
                                        schema=schema,
                                        remote_operation=True)[schema.backend]
        try:
            saved_digests, hashKVs = [], []
            for hdigest, tensor in received_data:
                hashVal = be_accessor.write_data(tensor, remote_operation=True)
                hashKey = hash_data_db_key_from_raw_key(hdigest)
                saved_digests.append(hdigest)
                hashKVs.append((hashKey, hashVal))
        finally:
            be_accessor.close()
        try:
            hashTxn = self.txnctx.begin_writer_txn(self.env.hashenv)
            for k, v in hashKVs:
                hashTxn.put(k, v)
        finally:
            self.txnctx.commit_writer_txn(self.env.hashenv)
        return saved_digests


RawCommitContent = NamedTuple('RawCommitContent', [('commit', str),
                                                   ('cmtParentVal', bytes),
                                                   ('cmtSpecVal', bytes),
                                                   ('cmtRefVal', bytes)])


class ContentReader(object):
    """Common methods to client & server which read content.

    These are special methods configured especially for remote operations.
    They do not honor the public facing API or data write/read conventions
    established for users or the rest of Hangar internals.

    Parameters
    ----------
    envs : context.Environments
        main hangar environment context object.
    """
    def __init__(self, envs):

        self.env: Environments = envs
        self.txnctx: TxnRegister = TxnRegister()

    def commit(self, commit: str) -> Union[RawCommitContent, bool]:
        """Read a commit with a given hash and get db formatted content

        Parameters
        ----------
        commit : str
            commit hash to read from the ref db

        Returns
        -------
        namedtuple or False
            nametuple with typename = RawCommitContent field_names = ('commit',
            'cmtParentVal', 'cmtSpecVal', 'cmtRefVal') if operation successful.

            False if commit does not exist with provided digest.
        """
        cmtRefKey = parsing.commit_ref_db_key_from_raw_key(commit)
        cmtParentKey = parsing.commit_parent_db_key_from_raw_key(commit)
        cmtSpecKey = parsing.commit_spec_db_key_from_raw_key(commit)

        reftxn = self.txnctx.begin_reader_txn(self.env.refenv)
        try:
            cmtRefVal = reftxn.get(cmtRefKey, default=False)
            cmtParentVal = reftxn.get(cmtParentKey, default=False)
            cmtSpecVal = reftxn.get(cmtSpecKey, default=False)
        finally:
            self.txnctx.abort_reader_txn(self.env.refenv)

        ret = RawCommitContent(commit, cmtParentVal, cmtSpecVal, cmtRefVal)

        if not all(ret) and not isinstance(ret.cmtParentVal, bytes):
            return False
        else:
            return ret

    def schema(self, schema_hash: str) -> Union[bytes, bool]:
        """Read db formatted schema val for a schema hash

        Parameters
        ----------
        schema_hash : str
            schema hash to look up

        Returns
        -------
        bytes or False
            db formatted representation of schema bytes if schema_hash exists

            False if the schema_hash does not exist in the db.
        """
        schemaKey = hash_schema_db_key_from_raw_key(schema_hash)
        hashTxn = self.txnctx.begin_reader_txn(self.env.hashenv)
        try:
            schemaVal = hashTxn.get(schemaKey, default=False)
        finally:
            self.txnctx.abort_reader_txn(self.env.hashenv)

        ret = False if not schemaVal else schemaVal
        return ret



from .descriptors import (
    Descriptor, OneOf, DictItems, EmptyDict, SizedIntegerTuple, checkedmeta
)
from .ndarray import NdarrayVariableShape, NdarrayFixedShape
from .pystring import StringVariableShape
from .pybytes import BytesVariableShape

__all__ = [
    'Descriptor', 'OneOf', 'DictItems', 'EmptyDict', 'SizedIntegerTuple',
    'checkedmeta', 'NdarrayVariableShape', 'NdarrayFixedShape',
    'StringVariableShape', 'BytesVariableShape'
]



import warnings
from typing import Iterable, Optional, Union, Tuple


ArraysetsRef = Union['ArraysetDataReader', Iterable['ArraysetDataReader']]


class GroupedColumns(object):
    """Groups hangar columns and validate suitability for usage in dataloaders.

    It can choose a subset of samples in the hangar columns by checking the
    list of keys or an index range. :class:`GroupedColumns` does not expect all
    the input hangar columns to have same length and same keys. It takes a
    `set.intersection` of sample names from all the columns and `keys`
    argument if passed and hence discard non-common keys. This is then
    intersected with samples which are available locally; not existing as
    remote partial data). Based on `keys` or `index_range` (ignore
    `index_range` if `keys` is present) it makes a subset of sample names which
    is then used to fetch the data from hangar columns.
    """

    def __init__(self,
                 columns: ArraysetsRef,
                 keys: Optional[Iterable[Union[int, str]]] = None,
                 index_range: Optional[slice] = None):

        self.columns_col = []
        self.column_names = []
        self._allowed_samples: Tuple[Union[str, int]] = None

        if not isinstance(columns, (list, tuple, set)):
            columns = (columns,)
        if len(columns) == 0:
            raise ValueError('len(columns) cannot == 0')

        column_lens = set()
        all_keys = []
        all_remote_keys = []
        for col in columns:
            if col.iswriteable is True:
                raise TypeError(f'Cannot load columns opened in `write-enabled` checkout.')
            self.columns_col.append(col)
            self.column_names.append(col.column)
            column_lens.add(len(col))
            all_keys.append(set(col.keys()))
            all_remote_keys.append(set(col.remote_reference_keys))

        if len(column_lens) > 1:
            warnings.warn('Columns do not contain equal number of samples', UserWarning)

        common_keys = set.intersection(*all_keys)
        remote_keys = set.union(*all_remote_keys)
        common_local_keys = common_keys.difference(remote_keys)

        if keys:
            keys = set(keys,)
            notCommon = keys.difference(common_keys)
            notLocal = keys.difference(common_local_keys)
            if len(notCommon) > 0:
                raise KeyError(f'Keys: {notCommon} do not exist in all columns.')
            if len(notLocal) > 0:
                raise FileNotFoundError(
                    f'Keys: {notLocal} are remote data samples not downloaded locally.')
            self._allowed_samples = tuple(keys)
        elif index_range:
            if not isinstance(index_range, slice):
                raise TypeError(f'type(index_range): {type(index_range)} != slice')
            # need to sort before slicing on index_range, but since sample keys
            # can be mixed int and str type, convert to common format and sort
            # on that first
            str_keys = [i if isinstance(i, str) else f'#{i}' for i in common_local_keys]
            sorted_keys = sorted(str_keys)
            converted_keys = [int(i[1:]) if i.startswith('#') else i for i in sorted_keys]
            self._allowed_samples = tuple(converted_keys)[index_range]
        else:
            self._allowed_samples = tuple(common_local_keys)

        if len(self._allowed_samples) == 0:
            raise ValueError(
                f'No Samples available common to all columns and available locally.')

    def get_types(self, converter=None):
        """
        Get dtypes of the all the columns in the `GroupedColumns`.

        Parameters
        ----------
        converter : Callable
            A function that takes default dtype (numpy) and convert it to another
            format

        Returns
        -------
        Tuple[np.dtype]
        """
        dtypes = []
        for col in self.columns_col:
            if converter:
                dtypes.append(converter(col.dtype))
            else:
                dtypes.append(col.dtype)
        return tuple(dtypes)

    def get_shapes(self, converter=None):
        """
        Get shapes of the all the columns in the `GroupedColumns`.

        Parameters
        ----------
        converter : Callable
            A function that takes default shape (numpy) and convert it to another
            format

        Returns
        -------
        A tuple of column shapes
        """
        shapes = []
        for col in self.columns_col:
            if col.schema_type == 'variable_shape':
                aset_shape = (None,) * len(col.shape)
            else:
                aset_shape = col.shape
            if converter:
                shapes.append(converter(aset_shape))
            else:
                shapes.append(aset_shape)
        return tuple(shapes)

    @property
    def sample_names(self):
        return self._allowed_samples



from typing import Iterable, Union, Tuple
import lmdb


class CursorRangeIterator:

    @staticmethod
    def cursor_range_iterator(datatxn: lmdb.Transaction, startRangeKey: bytes, keys: bool, values: bool
                              ) -> Iterable[Union[Tuple[bytes], Tuple[bytes, bytes]]]:
        """Common method used to implement cursor range iterators

        Parameters
        ----------
        datatxn : lmdb.Transaction
            open database transaction to read values from
        startRangeKey : bytes
            range in which to iterate cursor over until end of db or out of
            lexicographic range.
        keys : bool, optional
            If True, yield metadata keys encountered, if False only values
            are returned. By default, True.
        values : bool, optional
            If True, yield metadata hash values encountered, if False only
            keys are returned. By default, True.

        Yields
        ------
        Iterable[Union[Tuple[bytes], Tuple[bytes, bytes]]]:
            db keys or key/value tuple
        """
        len_RangeKey = len(startRangeKey)
        with datatxn.cursor() as cursor:
            rangeItemsExist = cursor.set_range(startRangeKey)
            if not rangeItemsExist:
                # break out prematurely in the case where no matching items exist.
                # Important to not disrupt callers who may expect to recieves some
                # iterable for processing.
                return iter([])

            # divide loop into returned type sections as perf optimization
            # (rather then if/else checking on every iteration of loop)
            if keys and not values:
                while rangeItemsExist:
                    recKey = cursor.key()
                    if recKey[:len_RangeKey] == startRangeKey:
                        yield recKey
                        rangeItemsExist = cursor.next()
                        continue
                    else:
                        rangeItemsExist = False
            elif values and not keys:
                while rangeItemsExist:
                    recKey, recVal = cursor.item()
                    if recKey[:len_RangeKey] == startRangeKey:
                        yield recVal
                        rangeItemsExist = cursor.next()
                        continue
                    else:
                        rangeItemsExist = False
            elif keys and values:
                while rangeItemsExist:
                    recKey, recVal = cursor.item()
                    if recKey[:len_RangeKey] == startRangeKey:
                        yield (recKey, recVal)
                        rangeItemsExist = cursor.next()
                        continue
                    else:
                        rangeItemsExist = False
            else:  # pragma: no cover
                raise RuntimeError(f'Internal hangar error while iterating cursor records for '
                                   f' {startRangeKey}. one of [`keys`, `values`] must be True.')



import configparser
import os
import shutil
import tempfile
import time
from contextlib import contextmanager, closing
from pathlib import Path

import lmdb

from .heads import (
    get_branch_head_commit,
    get_staging_branch_head,
    set_branch_head_commit,
    set_staging_branch_head,
)
from .parsing import (
    cmt_final_digest,
    commit_parent_db_key_from_raw_key,
    commit_parent_db_val_from_raw_val,
    commit_parent_raw_key_from_db_key,
    commit_parent_raw_val_from_db_val,
    commit_ref_db_key_from_raw_key,
    commit_ref_db_val_from_raw_val,
    commit_ref_raw_val_from_db_val,
    commit_spec_db_key_from_raw_key,
    commit_spec_db_val_from_raw_val,
    commit_spec_raw_val_from_db_val,
    DigestAndBytes,
)
from ..constants import (
    CONFIG_USER_NAME,
    DIR_DATA_REMOTE,
    DIR_DATA_STAGE,
    DIR_DATA_STORE,
    LMDB_SETTINGS,
    SEP_KEY,
)
from ..txnctx import TxnRegister

"""
Reading commit specifications and parents.
------------------------------------------
"""


def expand_short_commit_digest(refenv: lmdb.Environment, commit_hash: str) -> str:
    """Find the a full commit hash from a short version provided by the user

    Parameters
    ----------
    refenv : lmdb.Environment
        db where the commit references are stored
    commit_hash : str
        short commit hash to search for in the repository

    Returns
    -------
    str
        full commit hash if short maps to a unique digest in the repo history

    Raises
    ------
    KeyError
        If the short commit hash can reference two full commit digests
    KeyError
        if no expanded commit digest is found starting with the short version.
    """
    reftxn = TxnRegister().begin_reader_txn(refenv)
    commitParentStart = commit_parent_db_key_from_raw_key(commit_hash)
    with reftxn.cursor() as cursor:
        shortHashExists = cursor.set_range(commitParentStart)
        if shortHashExists is True:
            commitKey = cursor.key()
            commit_key = commit_parent_raw_key_from_db_key(commitKey)
            cursor.next()
            cursor.next()
            nextHashExist = cursor.next()
            if nextHashExist is False:
                return commit_key
            nextCommitKey = cursor.key()
            next_commit_key = commit_parent_raw_key_from_db_key(nextCommitKey)
            if next_commit_key.startswith(commit_hash) is True:
                raise KeyError(f'Non unique short commit hash: {commit_hash}')
            else:
                return commit_key
        else:
            raise KeyError(f'No matching commit hash found starting with: {commit_hash}')


def check_commit_hash_in_history(refenv, commit_hash):
    """Check if a commit hash exists in the repository history

    Parameters
    ----------
    refenv : lmdb.Environment
        refenv where the commit history is stored
    commit_hash : str
        hash of the commit to check for existence

    Returns
    -------
    bool
        True if exists, otherwise False
    """
    reftxn = TxnRegister().begin_reader_txn(refenv)
    try:
        commitParentKey = commit_parent_db_key_from_raw_key(commit_hash)
        commitParentVal = reftxn.get(commitParentKey, default=False)
        isCommitInHistory = True if commitParentVal is not False else False
    finally:
        TxnRegister().abort_reader_txn(refenv)
    return isCommitInHistory


def get_commit_spec(refenv, commit_hash):
    """Get the commit specifications of a particular hash.

    Parameters
    ----------
    refenv : lmdb.Environment
        refenv where the specs are stored
    commit_hash : str
        commit hash to query

    Returns
    -------
    namedtuple
        named tuple with all the commit specs included

    Raises
    ------
    ValueError
        if no commit exists with the provided hash
    """
    reftxn = TxnRegister().begin_reader_txn(refenv)
    try:
        parentCommitSpecKey = commit_spec_db_key_from_raw_key(commit_hash)
        parentCommitSpecVal = reftxn.get(parentCommitSpecKey, default=False)
    finally:
        TxnRegister().abort_reader_txn(refenv)

    if parentCommitSpecVal is False:
        raise ValueError(f'No commit exists with the hash: {commit_hash}')

    parentCommitSpec = commit_spec_raw_val_from_db_val(parentCommitSpecVal)
    return parentCommitSpec.user_spec


def get_commit_ancestors(refenv, commit_hash):
    """find the ancestors of a particular commit hash.

    Parameters
    ----------
    refenv : lmdb.Environment
        lmdb environment where the commit refs are stored
    commit_hash : string
        commit hash to find the ancestors for

    Returns
    -------
    namedtuple
        Namedtuple describing is_merge_commit, master_ancestor, &
        child_ancestor (in the even of merge commit)

    Raises
    ------
    ValueError
        if no commit exists with the provided hash
    """

    reftxn = TxnRegister().begin_reader_txn(refenv)
    try:
        parentCommitKey = commit_parent_db_key_from_raw_key(commit_hash)
        parentCommitVal = reftxn.get(parentCommitKey, default=False)
    finally:
        TxnRegister().abort_reader_txn(refenv)

    if parentCommitVal is False:
        raise ValueError(f'No commit exists with the hash: {commit_hash}')

    parentCommitAncestors = commit_parent_raw_val_from_db_val(parentCommitVal)
    return parentCommitAncestors.ancestor_spec


def get_commit_ancestors_graph(refenv, starting_commit):
    """returns a DAG of all commits starting at some hash pointing to the repo root.

    Parameters
    ----------
    refenv : lmdb.Environment
        lmdb environment where the commit refs are stored
    starting_commit : string
        commit hash to start creating the DAG from

    Returns
    -------
    dict
        a dictionary where each key is a commit hash encountered along the way,
        and it's value is a list containing either one or two elements which
        identify the child commits of that parent hash.
    """
    parent_commit = starting_commit
    commit_graph = {}
    seen = set(starting_commit)
    more_work = []
    end_commit = False

    if parent_commit == '':
        end_commit = True

    while end_commit is not True:
        childCommit = get_commit_ancestors(refenv, parent_commit)

        if (childCommit.master_ancestor == '') or (childCommit.master_ancestor in seen):
            end_commit = True
            commit_graph[parent_commit] = [childCommit.master_ancestor]
            if len(more_work) != 0:
                master_commit = more_work.pop(0)
                end_commit = False
            else:
                continue

        elif childCommit.is_merge_commit is True:
            master_commit = childCommit.master_ancestor
            dev_commit = childCommit.dev_ancestor
            more_work.append(dev_commit)
            commit_graph[parent_commit] = [master_commit, dev_commit]
            seen.add(master_commit)
            seen.add(dev_commit)

        else:
            master_commit = childCommit.master_ancestor
            commit_graph[parent_commit] = [master_commit]
            seen.add(master_commit)

        parent_commit = master_commit

    return commit_graph


"""
Methods for reading packed commit data and reconstructing an unpacked format.
-----------------------------------------------------------------------------
"""


def get_commit_ref(refenv, commit_hash):
    """Read the commit data record references from a specific commit.

    This only returns a list of tuples with binary encoded key/value pairs.

    Parameters
    ----------
    refenv : lmdb.Environment`
        lmdb environment where the references are stored
    commit_hash : string
        hash of the commit to retrieve.

    Returns
    -------
    tuple
        tuple of tuples containing encoded key/value pairs of the data
        records

    Raises
    ------
    ValueError
        if no commit exists with the provided hash
    """
    reftxn = TxnRegister().begin_reader_txn(refenv)
    try:
        cmtRefKey = commit_ref_db_key_from_raw_key(commit_hash)
        cmtSpecKey = commit_spec_db_key_from_raw_key(commit_hash)
        cmtParentKey = commit_parent_db_key_from_raw_key(commit_hash)

        cmtRefVal = reftxn.get(cmtRefKey, default=False)
        cmtSpecVal = reftxn.get(cmtSpecKey, default=False)
        cmtParentVal = reftxn.get(cmtParentKey, default=False)
    except lmdb.BadValsizeError:
        raise ValueError(f'No commit exists with the hash: {commit_hash}')
    finally:
        TxnRegister().abort_reader_txn(refenv)

    if (cmtRefVal is False) or (cmtSpecVal is False) or (cmtParentVal is False):
        raise ValueError(f'No commit exists with the hash: {commit_hash}')

    commitRefs = commit_ref_raw_val_from_db_val(cmtRefVal)
    commitSpecs = commit_spec_raw_val_from_db_val(cmtSpecVal)
    commitParent = commit_parent_raw_val_from_db_val(cmtParentVal)

    calculatedDigest = cmt_final_digest(
        parent_digest=commitParent.digest,
        spec_digest=commitSpecs.digest,
        refs_digest=commitRefs.digest)

    if calculatedDigest != commit_hash:
        raise IOError(
            f'Data Corruption Detected. On retrieval of stored references for '
            f'commit_hash: {commit_hash} validation of commit record/contents '
            f'integrity failed. Calculated digest: {calculatedDigest} != '
            f'expected: {commit_hash}. Please alert the Hangar development team to '
            f'this error if possible.')

    return commitRefs.db_kvs


def unpack_commit_ref(refenv, cmtrefenv, commit_hash):
    """unpack a commit record ref into a new key/val db for reader checkouts.

    This method also validates that the record data (parent, spec, and refs)
    have not been corrupted on disk (ie)

    Parameters
    ----------
    refenv : lmdb.Environment
        environment handle open for reading in the refenv
    cmtrefenv : lmdb.Environment
        environment handle open for writing on disk. this db must be empty.
    commit_hash : str
        hash of the commit to read in from refs and unpack in a checkout.
    """

    commitRefs = get_commit_ref(refenv=refenv, commit_hash=commit_hash)
    cmttxn = TxnRegister().begin_writer_txn(cmtrefenv)
    try:
        with cmttxn.cursor() as cursor:
            cursor.first()
            cursor.putmulti(commitRefs, append=True)
        try:
            cursor.close()
        except Exception as e:
            msg = f'could not close cursor cmttxn {cmttxn} commit_hash {commit_hash}'
            e.args = (*e.args, msg)
            raise e
    finally:
        TxnRegister().commit_writer_txn(cmtrefenv)

    return


@contextmanager
def tmp_cmt_env(refenv: lmdb.Environment, commit_hash: str):
    """create temporary unpacked lmdb environment from compressed structure

    Parameters
    ----------
    refenv : lmdb.Environment
        lmdb environment where the commit refs are stored
    commit_hash : str
        hash of the commit to get the contents of

    Returns
    -------
    lmdb.Environment
        environment with all db contents from ``commit`` unpacked
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpDF = os.path.join(tmpdir, 'test.lmdb')
        with closing(
                lmdb.open(tmpDF, sync=False, writemap=True, **LMDB_SETTINGS)
        ) as tmpDB:
            unpack_commit_ref(refenv, tmpDB, commit_hash)
            yield tmpDB


"""
Methods to write new commits
----------------------------

The functions below act to:
    - Reading and formatting all record data from the staging area.
    - Determining the ancestor(s) of the new commit
    - Specify commit details (message, time, committer-info, etc.)
    - Coordinate record hashing
    - Write the commit record
    - Update the branch head to point to the new commit hash
"""

# ---------------- Functions to format the written values of a commit --------------------


def _commit_ancestors(branchenv: lmdb.Environment,
                      *,
                      is_merge_commit: bool = False,
                      master_branch: str = '',
                      dev_branch: str = '') -> DigestAndBytes:
    """Format the commit parent db value, finding HEAD commits automatically.

    This method handles formatting for both regular & merge commits through the
    the keyword only arguments.

    Parameters
    ----------
    branchenv : lmdb.Environment
        Lmdb environment where branch data is located. If not merge commit, head
        branch and commit will be found.
    is_merge_commit : bool, optional
        If this is a merge commit or now, defaults to False
    master_branch : str, optional
        If merge commit, the master branch name must be specified, and the
        branch HEAD commit hash will be determined automatically, defaults to ''
    dev_branch : str, optional
        If merge commit, the dev branch name must be specified, and the branch
        HEAD commit hash will be determined automatically, defaults to ''

    Returns
    -------
    DigestAndBytes
        Commit parent db value and digest of commit parent val formatted
        appropriately based on the repo state and any specified arguments.
    """
    if not is_merge_commit:
        masterBranch = get_staging_branch_head(branchenv)
        master_ancestor = get_branch_head_commit(branchenv, masterBranch)
        dev_ancestor = ''
    else:
        master_ancestor = get_branch_head_commit(branchenv, master_branch)
        dev_ancestor = get_branch_head_commit(branchenv, dev_branch)

    commitParentVal = commit_parent_db_val_from_raw_val(
        master_ancestor=master_ancestor,
        dev_ancestor=dev_ancestor,
        is_merge_commit=is_merge_commit)

    return commitParentVal


def _commit_spec(message: str, user: str, email: str) -> DigestAndBytes:
    """Format the commit specification according to the supplied username and email.

    This method currently only acts as a pass through to the parsing options
    (with time filled in).

    Parameters
    ----------
    message : string
        Commit message sent in by the user.
    user : str, optional
        Name of the committer
    email : str, optional
        Email of the committer

    Returns
    -------
    DigestAndBytes
        Formatted value for the specification field of the commit and digest of
        spec.
    """
    spec_db = commit_spec_db_val_from_raw_val(commit_time=time.time(),
                                              commit_message=message,
                                              commit_user=user,
                                              commit_email=email)
    return spec_db


def _commit_ref(stageenv: lmdb.Environment) -> DigestAndBytes:
    """Query and format all staged data records, and format it for ref storage.

    Parameters
    ----------
    stageenv : lmdb.Environment
        lmdb environment where the staged record data is actually stored.

    Returns
    -------
    DigestAndBytes
        Serialized and compressed version of all staged record data along with
        digest of commit refs.
    """
    from .queries import RecordQuery  # needed to avoid cyclic import

    querys = RecordQuery(dataenv=stageenv)
    allRecords = tuple(querys._traverse_all_records())
    res = commit_ref_db_val_from_raw_val(allRecords)
    return res


# -------------------- Format ref k/v pairs and write the commit to disk ----------------


def commit_records(message, branchenv, stageenv, refenv, repo_path: Path,
                   *, is_merge_commit=False, merge_master=None, merge_dev=None):
    """Commit all staged records to the repository, updating branch HEAD as needed.

    This method is intended to work for both merge commits as well as regular
    ancestor commits.

    Parameters
    ----------
    message : string
        Message the user associates with what has been added, removed, or
        changed in this commit. Must not be empty.
    branchenv : lmdb.Environment
        lmdb environment where branch records are stored.
    stageenv : lmdb.Environment
        lmdb environment where the staged data records are stored in
        uncompressed format.
    refenv : lmdb.Environment
        lmdb environment where the commit ref records are stored.
    repo_path : Path
        path to the hangar repository on disk
    is_merge_commit : bool, optional
        Is the commit a merge commit or not? defaults to False
    merge_master : string, optional
        If merge commit, specify the name of the master branch, defaults to None
    merge_dev : string, optional
        If merge commit, specify the name of the dev branch, defaults to None

    Returns
    -------
    string
        Commit hash of the newly added commit
    """
    cmtParent = _commit_ancestors(branchenv=branchenv,
                                  is_merge_commit=is_merge_commit,
                                  master_branch=merge_master,
                                  dev_branch=merge_dev)

    user_info_pth = Path(repo_path, CONFIG_USER_NAME)
    CFG = configparser.ConfigParser()
    CFG.read(user_info_pth)

    USER_NAME = CFG['USER']['name']
    USER_EMAIL = CFG['USER']['email']
    if (USER_NAME is None) or (USER_EMAIL is None):
        raise RuntimeError(f'Username and Email are required. Please configure.')

    cmtSpec = _commit_spec(message=message, user=USER_NAME, email=USER_EMAIL)
    cmtRefs = _commit_ref(stageenv=stageenv)

    commit_hash = cmt_final_digest(parent_digest=cmtParent.digest,
                                   spec_digest=cmtSpec.digest,
                                   refs_digest=cmtRefs.digest)

    commitSpecKey = commit_spec_db_key_from_raw_key(commit_hash)
    commitParentKey = commit_parent_db_key_from_raw_key(commit_hash)
    commitRefKey = commit_ref_db_key_from_raw_key(commit_hash)

    reftxn = TxnRegister().begin_writer_txn(refenv)
    try:
        reftxn.put(commitSpecKey, cmtSpec.raw, overwrite=False)
        reftxn.put(commitParentKey, cmtParent.raw, overwrite=False)
        reftxn.put(commitRefKey, cmtRefs.raw, overwrite=False)
    finally:
        TxnRegister().commit_writer_txn(refenv)

    # possible separate function
    move_process_data_to_store(repo_path)
    if is_merge_commit is False:
        headBranchName = get_staging_branch_head(branchenv)
        set_branch_head_commit(branchenv, headBranchName, commit_hash)
    else:
        set_staging_branch_head(branchenv=branchenv, branch_name=merge_master)
        set_branch_head_commit(branchenv, merge_master, commit_hash)

    return commit_hash


# --------------------- staging setup, may need to move this elsewhere ------------------


def replace_staging_area_with_commit(refenv, stageenv, commit_hash):
    """DANGER ZONE: Delete the stage db and replace it with a copy of a commit environment.

    .. warning::

        In the current implementation, this method will not validate that it is safe
        to do this operation. All validation logic must be handled upstream.

    Parameters
    ----------
    refenv : [type]
        lmdb environment opened to the long term storage commit env
    stageenv : lmdb.Environment
        lmdb environment opened to the staging area.
    commit_hash : str
        commit hash to read from the refenv and replace the stage contents with.
    """
    stagetxn = TxnRegister().begin_writer_txn(stageenv)
    with stagetxn.cursor() as cursor:
        positionExists = cursor.first()
        while positionExists:
            positionExists = cursor.delete()
    cursor.close()
    TxnRegister().commit_writer_txn(stageenv)

    unpack_commit_ref(refenv=refenv, cmtrefenv=stageenv, commit_hash=commit_hash)
    return


def replace_staging_area_with_refs(stageenv, sorted_content):
    """DANGER ZONE: Delete all stage db records and replace it with specified data.

    .. warning::

        In the current implementation, this method will not validate that it is safe
        to do this operation. All validation logic must be handled upstream.

    Parameters
    ----------
    stageenv : lmdb.Environment
        staging area db to replace all data in.
    sorted_content : iterable of tuple
        iterable containing two-tuple of byte encoded record data to place in
        the stageenv db. index 0 -> db key; index 1 -> db val, it is assumed
        that the order of the tuples is lexicographically sorted by index 0
        values, if not, this will result in unknown behavior.
    """
    stagetxn = TxnRegister().begin_writer_txn(stageenv)
    with stagetxn.cursor() as cursor:
        positionExists = cursor.first()
        while positionExists:
            positionExists = cursor.delete()
    cursor.close()
    TxnRegister().commit_writer_txn(stageenv)

    cmttxn = TxnRegister().begin_writer_txn(stageenv)
    try:
        with cmttxn.cursor() as cursor:
            cursor.first()
            cursor.putmulti(sorted_content, append=True)
        cursor.close()
    finally:
        TxnRegister().commit_writer_txn(stageenv)


def move_process_data_to_store(repo_path: Path, *, remote_operation: bool = False):
    """Move symlinks to hdf5 files from process directory to store directory

    In process writes never directly access files in the data directory.
    Instead, when the file is created is is symlinked to either the remote data
    or stage data directory. All access is handled through this intermediate
    symlink in order to prevent any ability to overwrite (even if there are
    major errors in the hash records). Once the write operation is packed in
    the staging or remote area, this method is called to move the symlinks from
    the write enabled directory to the (read only, fully-committed) storage
    dir.

    Parameters
    ----------
    repo_path : Path
        path to the repository on dir
    remote_operation : bool, optional
        If this operation is occurring from a remote fetch operation. (the
        default is False, which means that all changes will occur in the
        staging area)

    """
    store_dir = Path(repo_path, DIR_DATA_STORE)

    type_dir = DIR_DATA_REMOTE if remote_operation else DIR_DATA_STAGE
    process_dir = Path(repo_path, type_dir)

    store_fps = []
    for be_pth in process_dir.iterdir():
        if be_pth.is_dir():
            for fpth in be_pth.iterdir():
                if fpth.is_file() and not fpth.stem.startswith('.'):
                    store_fps.append(store_dir.joinpath(be_pth.name, fpth.name))

    for fpth in store_fps:
        if not fpth.parent.is_dir():
            fpth.parent.mkdir()
        fpth.touch()

    # reset before releasing control.
    shutil.rmtree(process_dir)
    process_dir.mkdir(exist_ok=False)


def list_all_commits(refenv):
    """returns a list of all commits stored in the repository

    Parameters
    ----------
    refenv : lmdb.Environment
        db where all commit data is stored

    Returns
    -------
    list
        list of all commit digests.
    """
    refTxn = TxnRegister().begin_reader_txn(refenv)
    try:
        commits = set()
        with refTxn.cursor() as cursor:
            cursor.first()
            for k in cursor.iternext(keys=True, values=False):
                commitKey, *_ = k.decode().split(SEP_KEY)
                commits.add(commitKey)
            cursor.close()
    finally:
        TxnRegister().abort_reader_txn(refenv)

    return list(commits)



import pkg_resources
from typing import Callable


class PluginManager(object):
    """
    Container class that holds the information about available plugins and
    provides required method to fetch and clean up the plugin systems
    """

    valid_provides = ['load', 'save', 'show', 'board_show']

    def __init__(self):
        self._plugin_store = {}  # ex: {'pil': loaded_pil_module}
        self._default_plugins = {}  # ex: {'jpg': {'save': 'pil'}}
        self.plugins_loaded = False

    def reset_plugins(self):
        """
        Reset plugin clears the existing storages and then scans
        `hangar.external.plugins` for plugins. Once `plugin store` is populated,
        it creates finds the default plugins to make auto-inference based on
        file format possible
        """
        self._clear_plugins()
        self._scan_plugins()
        self._read_defaults()

    def _clear_plugins(self):
        """
        Clear the plugin state to the default, i.e., where no plugins are loaded
        """
        self._plugin_store.clear()
        self._default_plugins.clear()
        self.plugins_loaded = False

    def _scan_plugins(self):
        """
        Scan for entry points, find the plugins and store them in provided storage
        containers
        """
        for entry_point in pkg_resources.iter_entry_points('hangar.external.plugins'):
            PluginClass = entry_point.load()
            self._plugin_store[entry_point.name] = PluginClass()
        self.plugins_loaded = True

    def _read_defaults(self):
        """
        Populate default plugin dict that maps file formats to plugins and methods. This
        is used to infer which plugin to use at runtime based on file format
        """
        for fname, plugin in self._plugin_store.items():
            generator = ((ext, method) for ext in plugin.accepts for method in plugin.provides)
            for pair in generator:
                if pair not in self._default_plugins:
                    self._default_plugins[pair] = fname

    def get_plugin(self, method: str, plugin: str = None, extension: str = None) -> Callable:
        """Load installed plugin.

        User either needs to specify which plugin to load or should provide
        file format to infer which plugin to use

        Parameters
        ----------
        method : str
            Which method to import from the plugin. Methods implemented by the
            extension author should be declared as arguments passed into the
            BasePlugin superclass constructor
        plugin : str, optional
            Which plugin to load the method from. Cannot leave as ``None`` if
            ``extension`` is also ``None``
        extension : str, optional
            format of the data on the disk. This information is used to infer
            which plugin to use in case ``plugin`` is not provided explicitly.
            Cannot leave as ``None`` if ``plugin`` is also ``None``

        Returns
        -------
        plugin_method : function
            requested method from the plugin
        """

        if not plugin:
            if not extension:
                raise ValueError("Both `plugin` and `extension` cannot be empty together")

            plugin = self._default_plugins.get((extension, method))
            if plugin is None:
                raise ValueError(f"No plugins found for the file extension {extension} that could "
                                 f"do {method}")
        else:
            if plugin not in self._plugin_store:
                raise ValueError(f"Plugin {plugin} not found")
        loaded_plugin = self._plugin_store[plugin]
        try:
            return getattr(loaded_plugin, method)
        except AttributeError:
            raise RuntimeError(f"Method {method} found in `plugin.provides` but could "
                               f"not invoke from {plugin}. You might have forgot to define "
                               f"the function")



import logging
import tempfile
import time
import warnings
from collections import defaultdict
from contextlib import closing
from pathlib import Path
from typing import List, NamedTuple, Optional, Sequence

import grpc
import lmdb
from tqdm import tqdm
import numpy as np

from .backends import backend_decoder
from .constants import LMDB_SETTINGS
from .context import Environments
from .records import hash_data_db_key_from_raw_key
from .records import heads, queries, summarize
from .records.commiting import (
    check_commit_hash_in_history,
    move_process_data_to_store,
    unpack_commit_ref,
)
from .remote.client import HangarClient
from .remote.content import ContentWriter, ContentReader
from .txnctx import TxnRegister
from .utils import is_suitable_user_key

logger = logging.getLogger(__name__)

RemoteInfo = NamedTuple('RemoteInfo', [('name', str), ('address', str)])


class Remotes(object):
    """Class which governs access to remote interactor objects.

    .. note::

       The remote-server implementation is under heavy development, and is
       likely to undergo changes in the Future. While we intend to ensure
       compatability between software versions of Hangar repositories written
       to disk, the API is likely to change. Please follow our process at:
       https://www.github.com/tensorwerk/hangar-py

    """

    def __init__(self, env: Environments):

        self._env: Environments = env
        self._repo_path: Path = self._env.repo_path
        self._client: Optional[HangarClient] = None

    def __verify_repo_initialized(self):
        """Internal method to verify repo initialized before operations occur

        Raises
        ------
        RuntimeError
            If the repository db environments have not been initialized at the
            specified repo path.
        """
        if not self._env.repo_is_initialized:
            raise RuntimeError(
                f'Path {self._repo_path} not Hangar Repo. Use `init_repo()` method')

    def add(self, name: str, address: str) -> RemoteInfo:
        """Add a remote to the repository accessible by `name` at `address`.

        Parameters
        ----------
        name : str
            the name which should be used to refer to the remote server (ie:
            'origin')
        address : str
            the IP:PORT where the hangar server is running

        Returns
        -------
        RemoteInfo
            Two-tuple containing (``name``, ``address``) of the remote added to
            the client's server list.

        Raises
        ------
        ValueError
            If provided name contains any non ascii letter characters
            characters, or if the string is longer than 64 characters long.
        ValueError
            If a remote with the provided name is already listed on this client,
            No-Op. In order to update a remote server address, it must be
            removed and then re-added with the desired address.
        """
        self.__verify_repo_initialized()
        if (not isinstance(name, str)) or (not is_suitable_user_key(name)):
            raise ValueError(
                f'Remote name {name} of type: {type(name)} invalid. Must be '
                f'string with only alpha-numeric (or "." "_" "-") ascii characters. '
                f'Must be <= 64 characters long.')

        succ = heads.add_remote(self._env.branchenv, name=name, address=address)
        if succ is False:
            raise ValueError(f'No-Op: Remote named: {name} already exists.')
        return RemoteInfo(name=name, address=address)

    def remove(self, name: str) -> RemoteInfo:
        """Remove a remote repository from the branch records

        Parameters
        ----------
        name : str
            name of the remote to remove the reference to

        Raises
        ------
        KeyError
            If a remote with the provided name does not exist

        Returns
        -------
        str
            The channel address which was removed at the given remote name
        """
        self.__verify_repo_initialized()
        try:
            address = heads.remove_remote(branchenv=self._env.branchenv, name=name)
        except KeyError as e:
            raise e
        return RemoteInfo(name=name, address=address)

    def list_all(self) -> List[RemoteInfo]:
        """List all remote names and addresses recorded in the client's repository.

        Returns
        -------
        List[RemoteInfo]
            list of namedtuple specifying (``name``, ``address``) for each
            remote server recorded in the client repo.
        """
        self.__verify_repo_initialized()
        res = []
        names = heads.get_remote_names(self._env.branchenv)
        for name in names:
            address = heads.get_remote_address(self._env.branchenv, name)
            res.append(RemoteInfo(name=name, address=address))
        return res

    def ping(self, name: str) -> float:
        """Ping remote server and check the round trip time.

        Parameters
        ----------
        name : str
            name of the remote server to ping

        Returns
        -------
        float
            round trip time it took to ping the server after the connection was
            established and requested client configuration was retrieved

        Raises
        ------
        KeyError
            If no remote with the provided name is recorded.
        ConnectionError
            If the remote server could not be reached.
        """
        self.__verify_repo_initialized()
        address = heads.get_remote_address(branchenv=self._env.branchenv, name=name)
        self._client = HangarClient(envs=self._env, address=address)
        with closing(self._client) as client:
            client: HangarClient
            start = time.time()
            client.ping_pong()
            elapsed = time.time() - start
        return elapsed

    def fetch(self, remote: str, branch: str) -> str:
        """Retrieve new commits made on a remote repository branch.

        This is semantically identical to a `git fetch` command. Any new commits
        along the branch will be retrieved, but placed on an isolated branch to
        the local copy (ie. ``remote_name/branch_name``). In order to unify
        histories, simply merge the remote branch into the local branch.

        Parameters
        ----------
        remote : str
            name of the remote repository to fetch from (ie. ``origin``)
        branch : str
            name of the branch to fetch the commit references for.

        Returns
        -------
        str
            Name of the branch which stores the retrieved commits.
        """
        self.__verify_repo_initialized()
        address = heads.get_remote_address(self._env.branchenv, name=remote)
        self._client = HangarClient(envs=self._env, address=address)
        CW = ContentWriter(self._env)

        with closing(self._client) as client:
            client: HangarClient

            # ----------------- setup / validate operations -------------------

            try:
                cHEAD = heads.get_branch_head_commit(self._env.branchenv, branch)
            except ValueError:
                # branch does not exist on local client
                try:
                    s_branch = client.fetch_branch_record(branch)
                    sHEAD = s_branch.rec.commit
                except grpc.RpcError as rpc_error:
                    if rpc_error.code() == grpc.StatusCode.NOT_FOUND:
                        # branch does not exist on remote
                        logger.error(rpc_error.details())
                    raise rpc_error
            else:
                c_bhistory = summarize.list_history(
                    self._env.refenv, self._env.branchenv, branch_name=branch)
                try:
                    s_branch = client.fetch_branch_record(branch)
                    sHEAD = s_branch.rec.commit
                except grpc.RpcError as rpc_error:
                    if rpc_error.code() == grpc.StatusCode.NOT_FOUND:
                        # branch does not exist on remote
                        logger.error(rpc_error.details())
                    raise rpc_error

                # verify histories are intact and should be synced
                if sHEAD == cHEAD:
                    warnings.warn(f'NoOp:  {sHEAD} == client HEAD {cHEAD}', UserWarning)
                    return branch
                elif sHEAD in c_bhistory['order']:
                    warnings.warn(
                        f'REJECTED: remote HEAD: {sHEAD} behind local: {cHEAD}', UserWarning)
                    return branch

            # ------------------- get data ------------------------------------

            mCmtResponse = client.fetch_find_missing_commits(branch)
            m_cmts = mCmtResponse.commits
            for commit in tqdm(m_cmts, desc='fetching commit data refs'):
                mSchemaResponse = client.fetch_find_missing_schemas(commit)
                for schema in mSchemaResponse.schema_digests:
                    schema_hash, schemaVal = client.fetch_schema(schema)
                    CW.schema(schema_hash, schemaVal)
                # Record missing data hash digests (does not get data itself)
                m_hashes = client.fetch_find_missing_hash_records(commit)
                m_schema_hash_map = defaultdict(list)
                for digest, schema_hash in m_hashes:
                    m_schema_hash_map[schema_hash].append((digest, schema_hash))
                for schema_hash, received_data in m_schema_hash_map.items():
                    CW.data(schema_hash, received_data, backend='50')

            # Get missing commit reference specification
            for commit in tqdm(m_cmts, desc='fetching commit spec'):
                cmt, parentVal, specVal, refVal = client.fetch_commit_record(commit)
                CW.commit(cmt, parentVal, specVal, refVal)

            # --------------------------- At completion -----------------------

            # Update (or create) remote branch pointer with new HEAD commit
            fetchBranchName = f'{remote}/{branch}'
            try:
                heads.create_branch(
                    self._env.branchenv, name=fetchBranchName, base_commit=sHEAD)
            except ValueError:
                heads.set_branch_head_commit(
                    self._env.branchenv, branch_name=fetchBranchName, commit_hash=sHEAD)

            return fetchBranchName

    def fetch_data(self,
                   remote: str,
                   branch: str = None,
                   commit: str = None,
                   *,
                   column_names: Optional[Sequence[str]] = None,
                   max_num_bytes: int = None,
                   retrieve_all_history: bool = False) -> List[str]:
        """Retrieve the data for some commit which exists in a `partial` state.

        Parameters
        ----------
        remote : str
            name of the remote to pull the data from
        branch : str, optional
            The name of a branch whose HEAD will be used as the data fetch
            point. If None, ``commit`` argument expected, by default None
        commit : str, optional
            Commit hash to retrieve data for, If None, ``branch`` argument
            expected, by default None
        column_names : Optional[Sequence[str]]
            Names of the columns which should be retrieved for the particular
            commits, any columns not named will not have their data fetched
            from the server. Default behavior is to retrieve all columns
        max_num_bytes : Optional[int]
            If you wish to limit the amount of data sent to the local machine,
            set a `max_num_bytes` parameter. This will retrieve only this
            amount of data from the server to be placed on the local disk.
            Default is to retrieve all data regardless of how large.
        retrieve_all_history : Optional[bool]
            if data should be retrieved for all history accessible by the parents
            of this commit HEAD. by default False

        Returns
        -------
        List[str]
            commit hashes of the data which was returned.

        Raises
        ------
            ValueError
                if branch and commit args are set simultaneously.
            ValueError
                if specified commit does not exist in the repository.
            ValueError
                if branch name does not exist in the repository.
        """
        self.__verify_repo_initialized()
        address = heads.get_remote_address(branchenv=self._env.branchenv, name=remote)
        self._client = HangarClient(envs=self._env, address=address)
        CW = ContentWriter(self._env)

        # ----------------- setup / validate operations -----------------------

        if all([branch, commit]):
            raise ValueError(f'``branch`` and ``commit`` args cannot be set simultaneously')
        if branch is not None:
            cmt = heads.get_branch_head_commit(self._env.branchenv, branch_name=branch)
        else:
            cmt = commit
            cmtExist = check_commit_hash_in_history(self._env.refenv, commit)
            if not cmtExist:
                raise ValueError(f'specified commit: {commit} does not exist in the repo.')

        # --------------- negotiate missing data to get -----------------------

        if retrieve_all_history is True:
            if isinstance(max_num_bytes, int):
                raise ValueError(
                    f'setting the maximum number of bytes transferred and requesting '
                    f'all history are incompatible arguments.')
            else:
                hist = summarize.list_history(self._env.refenv, self._env.branchenv, commit_hash=cmt)
                commits = hist['order']
        else:
            commits = [cmt]

        with tempfile.TemporaryDirectory() as tempD:
            # share unpacked ref db between dependent methods
            tmpDF = Path(tempD, 'test.lmdb')
            tmpDB = lmdb.open(path=str(tmpDF), **LMDB_SETTINGS)
            try:
                allHashs = set()
                # all history argument
                for commit in tqdm(commits, desc='counting objects'):
                    with tmpDB.begin(write=True) as txn:
                        with txn.cursor() as curs:
                            notEmpty = curs.first()
                            while notEmpty:
                                notEmpty = curs.delete()
                    unpack_commit_ref(self._env.refenv, tmpDB, commit)
                    recQuery = queries.RecordQuery(tmpDB)

                    # handle column_names option
                    cmt_column_names = recQuery.column_names()
                    if column_names is None:
                        cmt_columns = cmt_column_names
                    else:
                        cmt_columns = [col for col in column_names if col in cmt_column_names]
                    for col in cmt_columns:
                        cmtData_hashs = recQuery.column_data_hashes(col)
                        allHashs.update(cmtData_hashs)
            finally:
                tmpDB.close()

        try:
            hashTxn = TxnRegister().begin_reader_txn(self._env.hashenv)
            m_schema_hash_map = defaultdict(list)
            for hashVal in allHashs:
                hashKey = hash_data_db_key_from_raw_key(hashVal.digest)
                hashRef = hashTxn.get(hashKey)
                be_loc = backend_decoder(hashRef)
                if be_loc.backend == '50':
                    m_schema_hash_map[be_loc.schema_hash].append(hashVal.digest)
        finally:
            TxnRegister().abort_reader_txn(self._env.hashenv)

        # -------------------- download missing data --------------------------

        total_nbytes_seen = 0
        total_data = sum(len(v) for v in m_schema_hash_map.values())
        with closing(self._client) as client, tqdm(total=total_data, desc='fetching data') as pbar:
            client: HangarClient  # type hint
            stop = False
            for schema in m_schema_hash_map.keys():
                hashes = set(m_schema_hash_map[schema])
                while (len(hashes) > 0) and (not stop):
                    ret = client.fetch_data(schema, hashes)
                    # max_num_bytes option
                    if isinstance(max_num_bytes, int):
                        for idx, r_kv in enumerate(ret):
                            try:
                                total_nbytes_seen += r_kv[1].nbytes
                            except AttributeError:
                                total_nbytes_seen += len(r_kv[1])
                            if total_nbytes_seen >= max_num_bytes:
                                ret = ret[0:idx]
                                stop = True
                                break
                    saved_digests = CW.data(schema, ret)
                    pbar.update(len(saved_digests))
                    hashes = hashes.difference(set(saved_digests))

        move_process_data_to_store(self._repo_path, remote_operation=True)
        return commits

    def push(self, remote: str, branch: str,
             *, username: str = '', password: str = '') -> str:
        """push changes made on a local repository to a remote repository.

        This method is semantically identical to a ``git push`` operation.
        Any local updates will be sent to the remote repository.

        .. note::

            The current implementation is not capable of performing a
            ``force push`` operation. As such, remote branches with diverged
            histories to the local repo must be retrieved, locally merged,
            then re-pushed. This feature will be added in the near future.

        Parameters
        ----------
        remote : str
            name of the remote repository to make the push on.
        branch : str
            Name of the branch to push to the remote. If the branch name does
            not exist on the remote, the it will be created
        username : str, optional, kwarg-only
            credentials to use for authentication if repository push restrictions
            are enabled, by default ''.
        password : str, optional, kwarg-only
            credentials to use for authentication if repository push restrictions
            are enabled, by default ''.

        Returns
        -------
        str
            Name of the branch which was pushed
        """
        self.__verify_repo_initialized()
        try:
            address = heads.get_remote_address(self._env.branchenv, name=remote)
            cHEAD = heads.get_branch_head_commit(self._env.branchenv, branch)
        except (KeyError, ValueError) as e:
            raise e from None

        CR = ContentReader(self._env)
        self._client = HangarClient(envs=self._env,
                                    address=address,
                                    auth_username=username,
                                    auth_password=password)

        # ----------------- setup / validate operations -------------------

        with closing(self._client) as client:
            client: HangarClient  # type hinting for development
            CR: ContentReader
            c_bhistory = summarize.list_history(refenv=self._env.refenv,
                                                branchenv=self._env.branchenv,
                                                branch_name=branch)
            try:
                s_branch = client.fetch_branch_record(branch)
            except grpc.RpcError as rpc_error:
                # Do not raise if error due to branch not existing on server
                if rpc_error.code() != grpc.StatusCode.NOT_FOUND:
                    raise rpc_error
            else:
                sHEAD = s_branch.rec.commit
                if sHEAD == cHEAD:
                    warnings.warn(
                        f'NoOp: server HEAD: {sHEAD} == client HEAD: {cHEAD}', UserWarning)
                    return branch
                elif (sHEAD not in c_bhistory['order']) and (sHEAD != ''):
                    warnings.warn(
                        f'REJECTED: server branch has commits not on client', UserWarning)
                    return branch

            # --------------- negotiate missing data to send -------------------

            try:
                # First push op verifies user permissions if push restricted (NOT SECURE)
                res = client.push_find_missing_commits(branch)
                m_commits = res.commits
            except grpc.RpcError as rpc_error:
                if rpc_error.code() == grpc.StatusCode.PERMISSION_DENIED:
                    raise PermissionError(f'{rpc_error.code()}: {rpc_error.details()}')
                else:
                    raise rpc_error

            m_schemas = set()
            m_schema_hashs = defaultdict(set)
            with tempfile.TemporaryDirectory() as tempD:
                tmpDF = Path(tempD, 'test.lmdb')
                tmpDB = lmdb.open(path=str(tmpDF), **LMDB_SETTINGS)
                for commit in tqdm(m_commits, desc='counting objects'):
                    # share unpacked ref db between dependent methods
                    with tmpDB.begin(write=True) as txn:
                        with txn.cursor() as curs:
                            notEmpty = curs.first()
                            while notEmpty:
                                notEmpty = curs.delete()
                    unpack_commit_ref(self._env.refenv, tmpDB, commit)
                    # schemas
                    schema_res = client.push_find_missing_schemas(commit, tmpDB=tmpDB)
                    m_schemas.update(schema_res.schema_digests)
                    # data hashs
                    m_cmt_schema_hashs = defaultdict(list)
                    mis_hashes_sch = client.push_find_missing_hash_records(commit, tmpDB=tmpDB)
                    for hsh, schema in mis_hashes_sch:
                        m_cmt_schema_hashs[schema].append(hsh)
                    for schema, hashes in m_cmt_schema_hashs.items():
                        m_schema_hashs[schema].update(hashes)
                tmpDB.close()

            # ------------------------- send data -----------------------------

            # schemas
            for m_schema in tqdm(m_schemas, desc='pushing schemas'):
                schemaVal = CR.schema(m_schema)
                if not schemaVal:
                    raise KeyError(f'no schema with hash: {m_schema} exists')
                client.push_schema(m_schema, schemaVal)
            # data
            total_data = sum([len(v) for v in m_schema_hashs.values()])
            with tqdm(total=total_data, desc='pushing data') as p:
                for dataSchema, dataHashes in m_schema_hashs.items():
                    client.push_data(dataSchema, dataHashes, pbar=p)
                    p.update(1)
            # commit refs
            for commit in tqdm(m_commits, desc='pushing commit refs'):
                cmtContent = CR.commit(commit)
                if not cmtContent:
                    raise KeyError(f'no commit with hash: {commit} exists')
                client.push_commit_record(commit=cmtContent.commit,
                                          parentVal=cmtContent.cmtParentVal,
                                          specVal=cmtContent.cmtSpecVal,
                                          refVal=cmtContent.cmtRefVal)

            # --------------------------- At completion -----------------------

            # update local remote HEAD pointer
            branchHead = heads.get_branch_head_commit(self._env.branchenv, branch)
            try:
                client.push_branch_record(branch, branchHead)
            except grpc.RpcError as rpc_error:
                # Do not raise if error due to branch not existing on server
                if rpc_error.code() != grpc.StatusCode.ALREADY_EXISTS:
                    logger.warning(f'CODE: {rpc_error.code()} DETAILS:{rpc_error.details()}')
                else:
                    raise rpc_error
            else:
                cRemoteBranch = f'{remote}/{branch}'
                if cRemoteBranch not in heads.get_branch_names(self._env.branchenv):
                    heads.create_branch(branchenv=self._env.branchenv,
                                        name=cRemoteBranch,
                                        base_commit=branchHead)
                else:
                    heads.set_branch_head_commit(branchenv=self._env.branchenv,
                                                 branch_name=cRemoteBranch,
                                                 commit_hash=branchHead)
            return branch
